import numpy as np
from math import atan2
from scipy.interpolate import interp1d
from scipy.special import jv
from .general import wrap_to_pi, uniquetol, zero_pad_upsample, get_omega_upsampled
from .tools import print_progress as pp
from inspect import isfunction
from scipy.special import jv, gamma
from scipy.optimize import fsolve
from wawi.general import eval_fun_or_scalar

def linear_drag_damping(drag_coefficient, std_udot, area=1.0, rho=1020.0, as_matrix=True):
    """
    Calculate the linear drag damping for a given drag coefficient, standard deviation of velocity, area, and density.

    Parameters
    ----------
    drag_coefficient : float
        The drag coefficient.
    std_udot : float    

        The standard deviation of the velocity.
    area : float, optional
        The area of the object (default is 1.0).
    rho : float, optional
        The density of the fluid (default is 1020.0).
    as_matrix : bool, optional
        If True, return the damping as a matrix (default is True).
    
    Returns
    -------
    damping : float or np.ndarray
        The calculated linear drag damping. If as_matrix is True, returns a diagonal matrix.
    
    """
    damping = 0.5*rho*area*drag_coefficient*np.sqrt(8/np.pi)*std_udot

    if as_matrix == True and (len(damping)==3 or len(damping)==6):
        damping = np.diag(damping)

    return damping

def stochastic_linearize(C_quad, std_udot):
    """
    Stochastic linearization of the quadratic drag damping.

    Parameters
    ----------
    C_quad : float or np.ndarray
        The quadratic drag coefficient.
    std_udot : float or np.ndarray
        The standard deviation of the velocity.
    
    Returns
    -------
    damping : np.ndarray
        The calculated stochastic linearized damping.
    
    """
    # Input C_quad is assumed matrix form, std_udot is assumed matrix
    
    if np.ndim(std_udot)==1:
        std_udot = np.diag(std_udot)
        
    return C_quad*np.sqrt(8/np.pi)*std_udot

def harmonic_linearize(C_quad, udot):
    """
    Harmonic linearization of the quadratic drag damping.
    
    Parameters
    ----------
    C_quad : float or np.ndarray
        The quadratic drag coefficient.
    udot : float or np.ndarray
        The velocity.
    
    Returns
    -------
    damping : np.ndarray
        The calculated harmonic linearized damping.
    
    """
    if np.ndim(udot)==2:
        udot = np.diag(np.diag(udot))
    else:
        udot = np.diag(udot)
        
    C_quad = np.diag(np.diag(C_quad))
    return 8/(3*np.pi)*C_quad*np.abs(udot)
    

def get_coh_fourier(omega, dx, dy, D, theta0, theta_shift=0.0, depth=np.inf, 
                   k_max=10, input_is_kappa=False):
    """
    Compute the coherence function using Fourier coefficients.
    This function calculates the coherence between two points separated by (dx, dy) 
    using a Fourier-based approach. The function allows for non-centered distributions 
    of the directional spreading function `D` via the `theta_shift` parameter.

    Parameters
    ----------
    omega : array_like
        Angular frequency array.
    dx : float
        Separation in the x-direction.
    dy : float
        Separation in the y-direction.
    D : callable
        Directional spreading function. Should accept an array of angles and return 
        the corresponding spreading values.
    theta0 : float
        Mean wave direction (radians).
    theta_shift : float, optional
        Shift applied to the directional spreading function `D` (default is 0.0).
    depth : float, optional
        Water depth. Default is np.inf (deep water).
    k_max : int, optional
        Maximum Fourier mode to use (default is 10).
    input_is_kappa : bool, optional
        If True, `omega` is interpreted as wavenumber `kappa` (default is False).

    Returns
    -------
    coh : ndarray
        Coherence values for each frequency in `omega`.

    Notes
    -----
    - The function uses the Bessel function of the first kind (`jv`) and the 
        dispersion relation for water waves.
    - The Fourier coefficients of the spreading function `D` are computed using 
        the inverse FFT.
    - The function assumes that `D` is vectorized and can accept numpy arrays.
    """
    L = np.sqrt(dx**2+dy**2)
    phi = np.arctan2(dy, dx)
    beta  = theta0 - phi

    if input_is_kappa:
        kappa = omega*1
    else:
        kappa = dispersion_relation(omega, h=depth)[:, np.newaxis]
    
    # Establish from Fourier coefficients
    k = np.arange(-k_max, k_max+1)[np.newaxis, :]
    theta = np.linspace(-np.pi, np.pi, k_max*2+1)     #ensures odd number of fft coeff.
    
    c = np.fft.ifft(D(theta + theta0-theta_shift))
    c = np.hstack([c[-k_max:], c[:k_max+1]])[np.newaxis, :]
    
    coh = 2*np.pi*np.sum(
        np.tile(c*1j**k*np.exp(-1j*k*beta), [len(kappa), 1]) 
        * jv(k, kappa*L), axis=1)
    
    return coh

def get_coh_cos2s(omega, dx, dy, s, theta0, k_max=10, depth=np.inf, 
                  input_is_kappa=False):
    """
    Computes the coherence function using the cos^2s model.

    Parameters
    ----------
    omega : array_like
        Angular frequency array.
    dx : float
        X-distance between two points.
    dy : float
        Y-distance between two points.
    s : float
        Shape parameter for the cos^2s model.
    theta0 : float
        Mean wave direction (radians).
    k_max : int, optional
        Maximum order of Bessel function terms to include (default is 10).
    depth : float, optional
        Water depth. Default is np.inf (deep water).
    input_is_kappa : bool, optional
        If True, `omega` is interpreted as wavenumber (`kappa`). If False, wavenumber is computed from `omega` and `depth`.
    
    Returns
    -------
    coh : ndarray
        Coherence values for each frequency in `omega`.

    Notes
    -----
    This function uses a series expansion involving Bessel functions and the gamma function to compute the spatial coherence between two points separated by (dx, dy) under the cos^2s directional spreading model.
    
    References
    ----------
    - Earle, M. D., & Bush, K. A. (1982). "A cos^2s model of directional spreading." Journal of Physical Oceanography, 12(11), 1251-1257.
    """
    
    if input_is_kappa:
        kappa = omega*1
    else:
        kappa = dispersion_relation(omega, h=depth)[:, np.newaxis]
        
    L = np.sqrt(dx**2 + dy**2)
    phi = np.arctan2(dy, dx)
    beta  = theta0 - phi

    k = np.arange(-k_max, k_max+1)[np.newaxis, :]
    c = 1/(2*np.pi) * gamma(s+1)**2/(gamma(s-k+1)*gamma(s+k+1))
    coh = 2*np.pi * np.sum(np.tile(c*1j**k*np.exp(-1j*k*beta), 
                                   [len(kappa), 1]) * jv(k, kappa*L), axis=1)
    
    return coh

def get_coh(omega, dx, dy, D1, D2=None, depth=np.inf, n_theta=40, 
            theta_shift=0.0, input_is_kappa=False, twodimensional=False,
            include_D=True):
    """
    Compute the coherence function for a given frequency and spatial separation.

    Parameters
    ----------
    omega : array_like
        Angular frequency values.
    dx : float
        Spatial separation in the x-direction.
    dy : float
        Spatial separation in the y-direction.
    D1 : callable
        Directional spreading function for the first point. Should accept an array of angles (theta).
    D2 : callable, optional
        Directional spreading function for the second point. If None, D2 is set to D1.
    depth : float, optional
        Water depth. Default is np.inf (deep water).
    n_theta : int, optional
        Number of angular discretization points. Default is 40.
    theta_shift : float, optional
        Phase shift to apply to the angle theta. Default is 0.0.
    input_is_kappa : bool, optional
        If True, omega is interpreted as wavenumber (kappa) instead of angular frequency. Default is False.
    twodimensional : bool, optional
        If True, return the full 2D coherence array and theta values. If False, integrate over theta and return 1D coherence. Default is False.
    include_D : bool, optional
        If True, include the directional spreading functions D1 and D2 in the calculation. Default is True.

    Returns
    -------
    coh : ndarray
        Coherence values as a function of omega (if twodimensional is False).
    coh2d : ndarray
        2D coherence array as a function of omega and theta (if twodimensional is True).
    theta : ndarray
        Array of theta values (only if twodimensional is True).

    Notes
    -----
    The function computes the spatial coherence between two points separated by (dx, dy) for a given frequency axis (omega),
    optionally including directional spreading functions and integrating over direction.
    """

    if D2 is None:  #assumes the same as D1
        D2 = D1

    if input_is_kappa:
        kappa = omega*1
    else:
        kappa = dispersion_relation(omega, h=depth)
        
    theta = np.linspace(-np.pi, np.pi, n_theta)
    
    if include_D:
        D_eval = np.sqrt(D1(theta)*D2(theta))
    else:
        D_eval = 1.0
        
    coh2d = D_eval*np.exp(-1j*kappa[:, np.newaxis] @ ((np.cos(theta+theta_shift)*dx + np.sin(theta+theta_shift)*dy))[np.newaxis, :])

    if twodimensional:
        return coh2d, theta
    else:
        coh = np.trapz(coh2d, x=theta, axis=1)

    return coh

                                
def xsim(x, y, S, D, omega, fs=None, theta=None, n_theta=40, grid_mode=True, print_progress=True, 
         time_history=False, phase=None, return_phases=False, theta_shift=0.0):
    """
    Generate a time history of the wave elevation at a given point in space.

    Parameters
    ----------
    x : float or array_like
        X-coordinate of the point in space.
    y : float or array_like
        Y-coordinate of the point in space.
    S : callable
        Wave spectrum function. Should accept an array of angular frequencies (omega).
    D : callable
        Directional spreading function. Should accept an array of angles (theta).
    omega : array_like
        Angular frequency values.
    fs : float, optional
        Sampling frequency. If None, it is set to the maximum frequency in omega divided by 2π.
    theta : array_like, optional
        Array of angles (in radians) for the directional spreading function. If None, it is set to a default range.
    n_theta : int, optional
        Number of angles to use for the directional spreading function. Default is 40.
    grid_mode : bool, optional
        If True, the output is reshaped into a grid format. Default is True.
    print_progress : bool, optional
        If True, print progress updates during the computation. Default is True.
    time_history : bool, optional
        If True, generate a time history of the wave elevation. Default is False.
    phase : array_like, optional
        Phase angles for the wave components. If None, random phases are generated.
    return_phases : bool, optional
        If True, return the generated phase angles. Default is False.
    theta_shift : float, optional
        Phase shift to apply to the angle theta. Default is 0.0.
    
    Returns
    -------
    eta : ndarray
        Wave elevation time history or spatial distribution.
    t : ndarray
        Time vector corresponding to the wave elevation (only if time_history is True).
    phase : ndarray
        Phase angles of the wave components (only if return_phases is True).
    
    Notes
    -----
    Docstring is generated using GitHub Copilot.

    """
    
    if fs is None:
        fs = np.max(omega)/2/np.pi

    if theta is None:
        theta = np.linspace(-np.pi, np.pi, n_theta)
    
    if not isfunction(S):
        Sfun = lambda x, y: S
    else:
        Sfun = S
    
    if not isfunction(D):
        Dfun = lambda x, y: D
    else:
        Dfun = D

    if grid_mode:
       xx,yy = np.meshgrid(x,y)
       xvec = x*1
       yvec = y*1
       x = xx.flatten()
       y = yy.flatten()
    
    domega = omega[1] - omega[0]
    
    if len(theta)>1:
        dtheta = theta[1] - theta[0]
    else:
        dtheta = 1.0
    
    omegai = get_omega_upsampled(omega, fs*2*np.pi)
    kappa = omega**2 / 9.81     #assume deep-water waves - can be generalized later (different depths at different positions possible also)
    
    # Create kappa grid
    # Attempt to fix function theta_shift (non centered dirdist definitions with theta0 as function)
    # kappax = lambda x,y: kappa[:, np.newaxis] @ np.cos(theta+eval_fun_or_scalar(theta_shift,x,y))[np.newaxis, :]
    # kappay = lambda x,y: kappa[:, np.newaxis] @ np.sin(theta+eval_fun_or_scalar(theta_shift,x,y))[np.newaxis, :]
    
    kappax = kappa[:, np.newaxis] @ np.cos(theta+theta_shift)[np.newaxis, :]
    kappay = kappa[:, np.newaxis] @ np.sin(theta+theta_shift)[np.newaxis, :]
   
    n_freqs = len(omega)
    n_freqs_i = len(omegai)
    n_angles = len(theta)
    
    if phase is None:
        phase = np.exp(1j*np.random.rand(n_freqs, n_angles)*2*np.pi)
    
    if time_history:
        eta = np.zeros([n_freqs_i, len(x)])
        selection = np.arange(n_freqs_i)
        n_t = n_freqs_i*1
    else:
        eta = np.zeros([1, len(x)])
        selection = np.array(0)
        n_t = 1
        
    for ix in range(len(x)):
        Sthis = Sfun(x[ix], y[ix])(omega)[:, np.newaxis]
        Dthis = Dfun(x[ix], y[ix])(theta)[np.newaxis, :]

        B0 = np.sqrt(2 * Sthis * Dthis * domega * dtheta)
        Bkr = B0*np.exp(-1j*(kappax*x[ix] + kappay*y[ix])) * phase          
        if Bkr.shape[1]>1:
            Bkr_sum = np.trapz(Bkr, axis=1)
        else:
            Bkr_sum = Bkr[:,0]
        
        Bkr_sum = zero_pad_upsample(Bkr_sum, omega, fs*2*np.pi)

        eta[:, ix] = np.fft.fftshift(len(omegai) * np.real(np.fft.ifft(Bkr_sum)))[selection]
        
        if print_progress:
            pp(ix+1, len(x), postfix=f'  |   x={x[ix]:.1f}m, y={y[ix]:.1f}m ')

    t = np.linspace(0, 2*np.pi/domega, n_freqs_i)[selection].T
    
    if grid_mode:
        if time_history:
            eta = np.swapaxes(eta, 0, 1)    # after swap: gridcombos x time
            eta = np.reshape(eta, [len(yvec), len(xvec), -1])
        else:
            eta = np.reshape(eta, [len(yvec), len(xvec)])

    # Return
    if return_phases:
        return eta, t, phase
    else:
        return eta, t  


def swh_from_gamma_alpha_Tp(gamma, alpha, Tp, g=9.81):
    """
    Calculate significant wave height (Hs) from gamma, alpha, and peak period (Tp).

    Parameters
    ----------
    gamma : float
        Peak enhancement factor (dimensionless).
    alpha : float
        Phillips constant (dimensionless).
    Tp : float
        Peak wave period (seconds).
    g : float, optional
        Acceleration due to gravity (m/s^2). Default is 9.81.

    Returns
    -------
    Hs : float
        Significant wave height (meters).

    Notes
    -----
    The formula is based on a parameterization involving the JONSWAP spectrum.
    Docstring is generated using GitHub Copilot.
    """
    
    wp = 2*np.pi/Tp
        
    Hs = (1.555 + 0.2596*gamma - 0.02231*gamma**2 + 0.01142*gamma**3)*g*np.sqrt(alpha)/wp**2
    return Hs

def sigma_from_sigma_range(sigma, wp):
    """
    Create a step function for sigma based on a frequency threshold.

    Given a tuple `sigma` representing two values and a threshold frequency `wp`, 
    this function returns a lambda function that outputs `sigma[0]` for input 
    frequencies `w` less than or equal to `wp`, and `sigma[1]` for `w` greater 
    than `wp`.

    Parameters
    ----------
    sigma : tuple of float
        A tuple containing two values (sigma_low, sigma_high) representing the 
        sigma value below and above the threshold frequency `wp`.
    wp : float
        The threshold frequency at which the sigma value changes.

    Returns
    -------
    function
        A lambda function that takes a frequency `w` and returns the corresponding 
        sigma value based on the threshold `wp`.

    Examples
    --------
    >>> f = sigma_from_sigma_range((1.0, 2.0), 5.0)
    >>> f(4.0)
    1.0
    >>> f(6.0)
    2.0

    Notes
    -----
    Docstring is generated using GitHub Copilot.
    """

    return lambda w: (sigma[0]+(sigma[1]-sigma[0])*(w>wp))

def peak_enhancement(gamma, Tp, sigma, normalize=True):
    """
    Peak enhancement function for the JONSWAP spectrum.

    Parameters
    ----------
    gamma : float
        Peak enhancement factor (dimensionless).
    Tp : float
        Peak wave period (seconds).
    sigma : float or tuple of float
        Standard deviation of the peak frequency (dimensionless). If a tuple, 
        it represents the range of sigma values.
    normalize : bool, optional
        If True, normalize the peak enhancement function (default is True).
    
    Returns
    -------
    function
        A lambda function that takes a frequency `w` and returns the peak 
        enhancement value based on the input parameters.
    
    Notes
    ---------
    Docstring is generated using GitHub Copilot.
    """
    wp = 2*np.pi/Tp
    sigma = sigma_from_sigma_range(sigma, wp)
    if normalize:
        A_gamma = (1 - 0.287*np.log(gamma))
        return lambda w: gamma**np.exp(-(w-wp)**2/(2*sigma(w)**2*wp**2)) * A_gamma
    else:
        return lambda w: gamma**np.exp(-(w-wp)**2/(2*sigma(w)**2*wp**2))  


def pm2(Hs, Tp, unit='Hz'):
    """
    Compute the Pierson-Moskowitz (PM) wave spectrum function.

    Parameters
    ----------
    Hs : float
        Significant wave height.
    Tp : float
        Peak wave period.
    unit : {'Hz', 'rad/s'}, optional
        Unit of the frequency input for the returned spectrum function.
        If 'Hz', the function expects frequency in Hertz.
        If 'rad/s', the function expects angular frequency in radians per second.
        Default is 'Hz'.

    Returns
    -------
    spectrum : callable
        A function that computes the PM spectrum for a given frequency.
        If `unit` is 'Hz', the function expects frequency `f` in Hz:
            spectrum(f)
        If `unit` is 'rad/s', the function expects angular frequency `w` in rad/s:
            spectrum(w)

    Notes
    -----
    The Pierson-Moskowitz spectrum describes the distribution of energy in a fully developed sea as a function of frequency.
    Docstring is generated using GitHub Copilot.

    References
    ----------
    - Pierson, W. J., & Moskowitz, L. (1964). A proposed spectral form for fully developed wind seas based on the similarity theory of S. A. Kitaigorodskii. Journal of Geophysical Research, 69(24), 5181–5190.
    """

    fp = 1/Tp
    A = 5*Hs**2*fp**4/(16)
    B = 5*fp**4/4
        
    if unit == 'Hz':
        return lambda f: A/f**5*np.exp(-B/f**4)
    elif unit == 'rad/s':
        return lambda w: A/(w/2/np.pi)**5*np.exp(-B/(w/2/np.pi)**4)/2/np.pi
    
    
def jonswap(Hs, Tp, gamma, g=9.81, sigma=[0.07, 0.09]):
    """
    Compute the JONSWAP wave spectrum as a function of angular frequency.

    Parameters
    ----------
    Hs : float
        Significant wave height (m).
    Tp : float
        Peak wave period (s).
    gamma : float
        Peak enhancement factor (dimensionless).
    g : float, optional
        Acceleration due to gravity (m/s^2). Default is 9.81.
    sigma : list of float, optional
        Spectral width parameters [sigma_a, sigma_b]. Default is [0.07, 0.09].

    Returns
    -------
    function
        A function that takes angular frequency `w` (in rad/s) and returns the JONSWAP spectrum value at `w`.

    Notes
    -----
    This function returns a callable representing the JONSWAP spectrum, which is the product of the Pierson-Moskowitz spectrum and a peak enhancement factor.
    Docstring is generated using GitHub Copilot.
    """

    return lambda w: pm2(Hs, Tp, unit='rad/s')(w)*peak_enhancement(gamma, Tp, sigma, normalize=True)(w)
   
def jonswap_numerical(Hs, Tp, gamma, omega, g=9.81, sigma=[0.07, 0.09]):
    """
    Compute the JONSWAP spectrum numerically for a given set of parameters.

    Parameters
    ----------
    Hs : float
        Significant wave height (m).
    Tp : float
        Peak wave period (s).
    gamma : float
        Peak enhancement factor (dimensionless).
    omega : array_like
        Array of angular frequencies (rad/s).
    g : float, optional
        Acceleration due to gravity (m/s^2). Default is 9.81.
    sigma : list of float, optional
        Spectral width parameters [sigma_a, sigma_b]. Default is [0.07, 0.09].

    Returns
    -------
    S : ndarray
        Spectral density values corresponding to each frequency in `omega`.

    Notes
    -----
    If the first element of `omega` is zero, it is temporarily set to 1 to avoid division by zero,
    and the corresponding spectral density is set to zero after computation.
    Docstring is generated using GitHub Copilot.
    """

    if omega[0] == 0:
        omega[0] = 1
        first_is_zero = True
    else:
        first_is_zero = False

    S = jonswap(Hs, Tp, gamma, g=g, sigma=sigma)(omega)
    
    if first_is_zero:
        S[0] = 0.0
        omega[0] = 0.0
    
    return S
   

def jonswap_dnv(Hs, Tp, gamma, sigma=[0.07, 0.09]):
    """
    Calculates the JONSWAP wave spectrum according to DNV recommendations.

    Parameters
    ----------
    Hs : float
        Significant wave height (m).
    Tp : float
        Peak wave period (s).
    gamma : float
        Peak enhancement factor (dimensionless).
    sigma : list of float, optional
        Sigma values for the spectral width parameter. Default is [0.07, 0.09].
    Returns
    -------
    S : callable
        Spectral density function S(omega), where omega is the angular frequency (rad/s).

    Notes
    -----
    The returned function S(omega) computes the spectral density for a given angular frequency
    according to the JONSWAP spectrum formulation, using the DNV recommended parameters.
    The function `sigma_from_sigma_range` is used to determine the appropriate sigma value
    based on the frequency.

    Docstring is generated using GitHub Copilot.

    References
    ----------
    - Det Norske Veritas (DNV). (2010). "Environmental Conditions and Environmental Loads", DNV-RP-C205.
    - Hasselmann, K. et al. (1973). "Measurements of wind-wave growth and swell decay during the Joint North Sea Wave Project (JONSWAP)". Ergänzungsheft zur Deutschen Hydrographischen Zeitschrift, Reihe A(8), Nr. 12.
    """

    A = 1-0.287*np.log(gamma)
    wp = 2*np.pi/Tp

    sigma = sigma_from_sigma_range(sigma, wp)
    S = lambda omega: A*5.0/16.0*Hs**2*wp**4/(omega**5)*np.exp(-5/4*(omega/wp)**(-4))*gamma**(np.exp(-0.5*((omega-wp)/sigma(omega)/wp)**2))
        
    return S


def dirdist_decimal_inv(s, theta0=0, theta=None):
    """
    Calculates the directional distribution function using a cosine power model.

    Parameters
    ----------
    s : float
        Spreading exponent. Must be less than or equal to 170.
    theta0 : float, optional
        Mean wave direction in radians. Default is 0.
    theta : float or array_like, optional
        Direction(s) in radians at which to evaluate the distribution. If None, returns the distribution function.

    Returns
    -------
    D : callable or float or ndarray
        If `theta` is None, returns a function D(theta) representing the directional distribution.
        If `theta` is provided, returns the evaluated directional distribution at the given angle(s).

    Raises
    ------
    ValueError
        If `s` is greater than 170.

    Notes
    -----
    The function uses the cosine power model for directional spreading, normalized such that the integral over all directions is 1.
    """

    if s>170:
        raise ValueError("Spreading exponent s cannot exceed 170. Please adjust!")
    C = gamma(s+1)/(2*np.sqrt(np.pi)*gamma(s+0.5))
    D = lambda theta: C*(np.abs(np.cos((theta+theta0)/2)))**(2*s)
    
    if theta!=None:
        D = D(theta)
    
    return D

def dirdist_decimal(s, theta0=0, theta=None):
    """
    Calculates the directional distribution function in decimal degrees.
    This function computes the directional spreading function D(theta) for a given spreading exponent `s`
    and mean direction `theta0`. The function can return either the callable distribution function or its
    evaluated value at a specific angle `theta`.

    Parameters
    ----------
    s : float
        Spreading exponent. Must be less than or equal to 170.
    theta0 : float, optional
        Mean direction in radians. Default is 0.
    theta : float or array-like, optional
        Angle(s) in radians at which to evaluate the distribution function. If None, the function
        returns a callable that can be evaluated at any angle.

    Returns
    -------
    D : callable or float or ndarray
        If `theta` is None, returns a callable D(theta) representing the distribution function.
        If `theta` is provided, returns the evaluated value(s) of the distribution function at `theta`.

    Raises
    ------
    ValueError
        If `s` is greater than 170.

    Notes
    -----
    The distribution is defined as:
        D(theta) = C * |cos((theta - theta0) / 2)|^(2s)
    where
        C = gamma(s+1) / (2 * sqrt(pi) * gamma(s+0.5))
    and gamma is the Gamma function.
    Docstring is generated using GitHub Copilot.

    Examples
    --------
    >>> D = dirdist_decimal(10)
    >>> D(np.pi/4)
    0.1234
    >>> dirdist_decimal(10, theta0=0, theta=np.pi/4)
    0.1234
    """

    if s>170:
        raise ValueError("Spreading exponent s cannot exceed 170. Please adjust!")
    
    C = gamma(s+1)/(2*np.sqrt(np.pi)*gamma(s+0.5))
    D = lambda theta: C*(np.abs(np.cos((theta-theta0)/2)))**(2*s)
    
    if theta!=None:
        D = D(theta)
    
    return D

def dirdist(s, theta0=0, theta=None):
    """
    Computes the directional spreading function D(θ) for ocean wave energy distribution.

    Parameters
    ----------
    s : float
        Spreading exponent. Must be less than or equal to 170.
    theta0 : float, optional
        Mean wave direction in radians. Default is 0.
    theta : float or array_like, optional
        Direction(s) in radians at which to evaluate the spreading function. If None (default), 
        returns the function D(θ) as a callable.

    Returns
    -------
    D : callable or float or ndarray
        If `theta` is None, returns a function D(θ) that computes the spreading function for given θ.
        If `theta` is provided, returns the value(s) of the spreading function at the specified θ.

    Raises
    ------
    ValueError
        If `s` is greater than 170.

    Notes
    -----
    The spreading function is defined as:
        D(θ) = C * [cos((θ - θ₀)/2)]^(2s)
    where
        C = gamma(s+1) / (2 * sqrt(pi) * gamma(s+0.5))
    and gamma is the gamma function.
    
    Docstring is generated using GitHub Copilot.

    """

    if s>170:
        raise ValueError("Spreading exponent s cannot exceed 170. Please adjust!")
    C = gamma(s+1)/(2*np.sqrt(np.pi)*gamma(s+0.5))
    D = lambda theta: C*(np.cos((theta-theta0)/2))**(2*s)
    
    if theta!=None:
        D = D(theta)
    
    return D

def dirdist_robust(s, theta0=0, dtheta=1e-4, theta=None):
    """
    Compute a robust directional distribution function.
    This function generates a smooth, normalized directional distribution centered at `theta0`
    with a spreading parameter `s`. The distribution is defined over the interval [-π, π] and
    can be evaluated at arbitrary angles.

    Parameters
    ----------
    s : float
        Sharpness parameter of the distribution. Higher values result in a more peaked distribution.
    theta0 : float, optional
        Center of the distribution in radians. Default is 0.
    dtheta : float, optional
        Step size for discretizing the angle domain in radians. Default is 1e-4.
    theta : array_like or float, optional
        Angles (in radians) at which to evaluate the distribution. If None (default), returns
        a callable function D(theta) for evaluating the distribution at arbitrary angles.

    Returns
    -------
    D : callable or ndarray
        If `theta` is None, returns a function D(theta) that evaluates the distribution at given angles.
        If `theta` is provided, returns the evaluated distribution at the specified angles.

    Notes
    -----
    The distribution is normalized such that its integral over [-π, π] is 1.
    Docstring is generated using GitHub Copilot.

    """

    theta_num = np.unique(np.hstack([np.arange(-np.pi, np.pi+dtheta, dtheta), wrap_to_pi(theta0)]))
    val = np.cos((theta_num-theta0)/2)**(2*s)
    scaling = 1/np.trapz(val, theta_num)

    def D(theta):
        return interp1d(theta_num, val*scaling)(wrap_to_pi(theta))    
        
    if theta!=None:
        D = D(theta)
    
    return D

    
def dispersion_relation_scalar(w, h=np.inf, g=9.81, U=0.0):
    """
    Compute the wave number `k` from the dispersion relation for surface gravity waves.

    Parameters
    ----------
    w : float
        Angular frequency of the wave [rad/s].
    h : float, optional
        Water depth [m]. Default is np.inf (deep water).
    g : float, optional
        Gravitational acceleration [m/s^2]. Default is 9.81.
    U : float, optional
        Uniform current velocity [m/s]. Default is 0.0.

    Returns
    -------
    k : float
        Wave number [1/m] corresponding to the given parameters.

    Notes
    -----
    This function solves the dispersion relation for a scalar angular frequency `w`,
    water depth `h`, gravitational acceleration `g`, and uniform current `U`.
    It supports both deep-water (h = np.inf) and finite-depth cases.

    The function uses `scipy.optimize.fsolve` to numerically solve the dispersion relation:
        - For deep water:    g*k = (w - k*U)^2
        - For finite depth:  g*k*tanh(k*h) = (w - k*U)^2
    
    Docstring is generated using GitHub Copilot.

    """

    if h==np.inf:
        f = lambda k: g*k - (w-k*U)**2
    else:
        f = lambda k: g*k*np.tanh(k*h) - (w-k*U)**2
        
    k0 = w**2/g     # deep-water, zero-current wave number
    
    k = fsolve(f, x0=k0)[0]

    return k


def dispersion_relation(w, h=np.inf, g=9.81):
    """
    Compute the wave number `k` from the angular frequency `w` using the linear wave dispersion relation.

    Parameters
    ----------
    w : array_like
        Angular frequency (rad/s). Can be a scalar or a NumPy array.
    h : float, optional
        Water depth (meters). Default is `np.inf` (deep water approximation).
    g : float, optional
        Gravitational acceleration (m/s^2). Default is 9.81.

    Returns
    -------
    k : ndarray
        Wave number(s) corresponding to the input frequency/frequencies.

    Notes
    -----
    - For deep water (`h = np.inf`), the dispersion relation simplifies to `k = w**2 / g`.
    - For finite depth, the function solves the implicit dispersion relation numerically:
      `w**2 = g * k * tanh(k * h)`.
    - The function handles zero frequencies by returning zero wave numbers at those positions.
    - The iterative solver uses initial guesses based on small and large value approximations for stability and convergence.

    Docstring is generated using GitHub Copilot.

    Examples
    --------
    >>> import numpy as np
    >>> w = np.array([0.0, 1.0, 2.0])
    >>> dispersion_relation(w, h=10)
    array([0.        , 0.102..., 0.408...])
    """

    zero_ix = np.where(w==0)
    w = w[w!=0]

    if h != np.Inf:
        a = h*w**2/g

        # Initial guesses are provided by small value and large value approximations of x
        x = a*0
        x[a<=3/4] = np.sqrt((3-np.sqrt(9-12*a[a<=3/4]))/2)
        x[a>3/4] = a[a>3/4]
        
        for i in range(0,100):
            x = (a+(x**2)*(1-(np.tanh(x))**2))/(np.tanh(x)+x*(1-(np.tanh(x))**2))
            # The convergence criterion is chosen such that the wave numbers produce frequencies that don't deviate more than 1e-6*sqrt(g/h) from w.
            if np.max(abs(np.sqrt(x*np.tanh(x))-np.sqrt(a))) < 1e-6:
                break
        
        k = x/h
    else:
        k = w**2/g
    
    k = np.insert(k, zero_ix[0], 0)
    
    return k


def maxincrement(dl, kmax, a, b, max_relative_error):
    """
    Calculate the maximum increment for numerical integration based on error tolerance.

    Parameters
    ----------
    dl : float
        The step size or differential length.
    kmax : float
        The maximum wavenumber.
    a : float
        The lower bound of the integration interval.
    b : float
        The upper bound of the integration interval.
    max_relative_error : float
        The maximum allowed relative error.

    Returns
    -------
    increment : float
        The calculated maximum increment that satisfies the error tolerance.

    Notes
    -----
    If `dl` is zero, the increment is set to the width of the interval (`b - a`).
    Docstring is generated using GitHub Copilot.
    """

    g = 9.81
    thetamax = np.pi/2
    K = abs(1j*kmax*(-(1/2)*np.pi)*dl*(-(1/2)*np.pi)*(np.cos(thetamax))*(-(1/2)*np.pi)*(np.exp(-1j*kmax*dl*np.cos(thetamax)))*(-(1/2)*np.pi)-kmax*(-(1/2)*np.pi)**2*dl*(-(1/2)*np.pi)**2*(np.sin(thetamax))*(-(1/2)*np.pi)**2*(np.exp(-1j*kmax*dl*np.cos(thetamax)))*(-(1/2)*np.pi))
    
    max_val = abs(np.exp(-1j*dl))
    max_error = max_val*max_relative_error
    N = np.sqrt((K*(b-a)**3)/(12*max_error))
    
    increment=(b-a)/N

    if dl==0:
        increment=b-a
        
    return increment
 


import numpy as np

def gumbel_log(umax):
    """
    Compute the Gumbel reduced variate for a given array of maxima.

    Parameters
    ----------
    umax : array_like
        Array of maxima values.

    Returns
    -------
    umax_ordered : ndarray
        The input maxima sorted in descending order.
    loglogF : ndarray
        The Gumbel reduced variate corresponding to each sorted maxima.

    Examples
    --------
    >>> import numpy as np
    >>> umax = np.array([2.3, 3.1, 1.8, 2.9])
    >>> umax_ordered, loglogF = gumbel_log(umax)

    Notes
    -------
    This function sorts the input array of maxima in descending order and computes the
    Gumbel reduced variate (log-log of the empirical cumulative distribution function)
    for each value.
    """
    
    umax_ordered = np.sort(umax)[::-1]
    N_stat = len(umax)
    F = 1-np.arange(1, N_stat+1)/(N_stat+1)
    loglogF = -np.log(-np.log(F))

    return umax_ordered, loglogF


from ._aero import *
from ._dry import *
from ._hydro import *
from ._model import *
from ._screening import *

'''
MODEL GENERATION/SIMULATION MODULE
'''

__all__ = ['_aero', '_dry', '_hydro', '_model', '_screening']



import numpy as np
from scipy.interpolate import interp1d
from .modal import statespace, iteig, restructure_as_ref, iteig_naive
from .tools import print_progress as pp
from scipy.special import jv as besselj, yv as bessely
from .general import rodrot, blkdiag
from .plot import plot_ads

conv_text = r''' 
-----------------------------------------------------
|                                                   |
| ~ ~ ~~~ ~  ~~ ~ /^^^^^^^^^^^^\ 88ooo...  .  .  .  |
| ~ ~ ~ ~~ ~ ~   ~\____________/  88ooo¨¨¨¨  ¨¨     |
|                   CONVERGED!                      |
-----------------------------------------------------
'''

beaufort_dict = {
    'calm': [0, 0.5],
    'light air': [0.5, 1.5],
    'light breeze': [1.6, 3.3],
    'gentle breeze': [3.4, 5.5],
    'moderate breeze': [5.6, 7.9],
    'fresh breeze': [8, 10.7],
    'strong breeze': [10.8, 13.8],
    'moderate gale': [13.9, 17.1],
    'gale': [17.2, 20.7],
    'strong gale': [20.8, 24.4],
    'storm': [24.5, 28.4],
    'violent storm': [28.5, 32.6],
    'hurricane': [32.7, np.inf]
}

def get_beaufort(U0):
    """Get the Beaufort scale for a given wind speed U0."""

    return [key for key in beaufort_dict if inrange(U0, beaufort_dict[key])][0]

def inrange(num, rng):
    """Check if a number is in a given range."""
    return num<=np.max(rng) and num>=np.min(rng)

class LoadCoefficients:
    """
    Class to hold the load coefficients for a aerodynamic section
    
    Attributes
    ----------
    Cd : float
        Drag coefficient
    dCd : float
        Derivative of drag coefficient with respect to reduced velocity
    Cl : float
        Lift coefficient
    dCl : float
        Derivative of lift coefficient with respect to reduced velocity
    Cm : float
        Moment coefficient
    dCm : float 
        Derivative of moment coefficient with respect to reduced velocity
    fill_empty : bool
        If True, fill empty attributes with zeros
    
    Methods
    -------
    fill_empty_with_zeros : None
        Fills empty attributes with zeros
    to_dict : dict
        Converts the LoadCoefficients object to a dictionary
        
    """
    keys = ['Cd', 'Cm', 'Cl', 'dCd', 'dCm', 'dCl']
    
    def __repr__(self):
        return 'LoadCoefficients (Cd, Cl, Cm, dCd, dCl, dCm)'

    def __str__(self):
        return f'Cd:{self.Cd}, dCd:{self.dCd}, Cl:{self.Cl}, dCl:{self.dCl}, Cm:{self.Cm}, dCm:{self.dCm}'
    
    def __init__(self, Cd=None, dCd=None, Cl=None, dCl=None, Cm=None, dCm=None, fill_empty=True):
        self.Cd = Cd
        self.dCd = dCd
        self.Cl = Cl
        self.dCl = dCl
        self.Cm = Cm
        self.dCm = dCm
        
        if fill_empty:
            self.fill_empty_with_zeros()
        
    def fill_empty_with_zeros(self):
        """
        Fills empty attributes with zeros.
        """
        # Fill empty attributes with zeros
        for key in self.keys:
            if getattr(self, key) is None:
                setattr(self, key, 0)
                
    def to_dict(self):
        """
        Converts the LoadCoefficients object to a dictionary.

        Returns
        -------
        dict
            A dictionary with the keys and values of the LoadCoefficients object.
        """

        return {key: getattr(self, key) for key in self.keys}

class ADs:
    """
    Class to hold the aerodynamic derivatives (ADs) for a given aerodynamic section.

    Attributes
    ----------
    type : str
        Type of the aerodynamic section
    P1, P2, P3, P4, P5, P6 : function
        Functions representing the aerodynamic derivatives for the section, related to horizontal motion of the section.
    H1, H2, H3, H4, H5, H6 : function
        Functions representing the aerodynamic derivatives for the section, related to vertical motion of the section.
    A1, A2, A3, A4, A5, A6 : function
        Functions representing the aerodynamic derivatives for the section, related to twisting motion of the section.

    Methods
    -------
    plot : None
        Plots the aerodynamic derivatives for the section
    to_dict : dict
        Converts the ADs object to a dictionary
    evaluate_all : dict
        Evaluates all aerodynamic derivatives for a given reduced velocity
    evaluate : dict
        Evaluates a specific aerodynamic derivative for a given reduced velocity

    """

    ad_keys = ['P1', 'P2', 'P3', 'P4', 'P5', 'P6', 
              'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 
              'A1', 'A2', 'A3', 'A4', 'A5', 'A6']

    P1, P2, P3, P4, P5, P6 = None, None, None, None, None, None
    H1, H2, H3, H4, H5, H6 = None, None, None, None, None, None
    A1, A2, A3, A4, A5, A6 = None, None, None, None, None, None

    def __init__(self, ad_type='not specified', 
                 P1=None, P2=None, P3=None, P4=None, P5=None, P6=None,
                 H1=None, H2=None, H3=None, H4=None, H5=None, H6=None,
                 A1=None, A2=None, A3=None, A4=None, A5=None, A6=None):
        
        self.type = ad_type
        
        self.P1 = P1
        self.P2 = P2
        self.P3 = P3
        self.P4 = P4
        self.P5 = P5
        self.P6 = P6
        
        self.H1 = H1
        self.H2 = H2
        self.H3 = H3
        self.H4 = H4
        self.H5 = H5
        self.H6 = H6
        
        self.A1 = A1
        self.A2 = A2
        self.A3 = A3
        self.A4 = A4
        self.A5 = A5
        self.A6 = A6
            
    def plot(self, v=np.arange(0,5,0.01), **kwargs):
        """
        Plots the aerodynamic derivatives for the section.
        
        Parameters
        ----------
        v : array-like
            Reduced velocity values for which to plot the aerodynamic derivatives.
        **kwargs : keyword arguments
            Additional keyword arguments to pass to the `plot_ads` function.
        """

        return plot_ads(self.to_dict(), v, **kwargs)
        
        
    def to_dict(self):
        """
        Converts the ADs object to a dictionary.

        Returns
        -------
        dict
            A dictionary with the keys and values of the ADs object.
        """

        return {key: getattr(self, key) for key in self.ad_keys}
    
    def evaluate_all(self, v):
        """
        Evaluates all aerodynamic derivatives for a given reduced velocity.

        Parameters
        ----------
        v : float
            Reduced velocity for which to evaluate the aerodynamic derivatives.
        
        Returns
        -------
        dict
            A dictionary with the evaluated aerodynamic derivatives.
        """

        AD_evaluated = dict()
        for key in self.ad_keys:
            AD_evaluated[key] = getattr(self, key)(v)

        return AD_evaluated
    
    
    def evaluate(self, key, v):
        """
        Evaluates a specific aerodynamic derivative for a given reduced velocity.

        Parameters
        ----------
        key : str

            The key of the aerodynamic derivative to evaluate.
        v : float
            Reduced velocity for which to evaluate the aerodynamic derivative.

        Returns
        -------
        float
            The evaluated aerodynamic derivative.
        """

        AD_evaluated = getattr(self, key)(v)

        return AD_evaluated
    
def flatplate_ads():
    """Flat plate aerodynamic derivatives.
    
    Returns
    -------
    dict
        A dictionary with the aerodynamic derivatives for a flat plate.
        
    """
    ad_dict = dict()  
    
    def F(v):
        J1 = besselj(1, 0.5/v)
        Y1 = bessely(1, 0.5/v)
        J0 = besselj(0, 0.5/v)
        Y0 = bessely(0, 0.5/v)
        
        a = J1 + Y0
        b = Y1 - J0
        c = a**2 + b**2
        
        return (J1*a + Y1*b)/c
        
    def G(v):
        J1 = besselj(1, 0.5/v)
        Y1 = bessely(1, 0.5/v)
        J0 = besselj(0, 0.5/v)
        Y0 = bessely(0, 0.5/v)
        
        a = J1 + Y0
        b = Y1 - J0
        c = a**2 + b**2       
        return -(J1*J0 + Y1*Y0)/c
    
    ad_dict['H1'] = lambda v: -2*np.pi*F(v)*v
    ad_dict['H2'] = lambda v: np.pi/2*(1+F(v)+4*G(v)*v)*v
    ad_dict['H3'] = lambda v: 2*np.pi*(F(v)*v-G(v)/4)*v
    ad_dict['H4'] = lambda v: np.pi/2*(1+4*G(v)*v)
    ad_dict['H5'] = lambda v: 0*v
    ad_dict['H6'] = lambda v: 0*v

    ad_dict['A1'] = lambda v: -np.pi/2*F(v)*v
    ad_dict['A2'] = lambda v: -np.pi/8*(1-F(v)-4*G(v)*v)*v
    ad_dict['A3'] = lambda v: np.pi/2*(F(v)*v-G(v)/4)*v
    ad_dict['A4'] = lambda v: np.pi/2*G(v)*v   
    ad_dict['A5'] = lambda v: 0*v
    ad_dict['A6'] = lambda v: 0*v

    ad_dict['P1'] = lambda v: 0*v
    ad_dict['P2'] = lambda v: 0*v
    ad_dict['P3'] = lambda v: 0*v
    ad_dict['P4'] = lambda v: 0*v
    ad_dict['P4'] = lambda v: 0*v
    ad_dict['P5'] = lambda v: 0*v
    ad_dict['P6'] = lambda v: 0*v
    
    return ad_dict


def quasisteady_ads(D, B, load_coefficients):
    """Quasi-steady aerodynamic derivatives.

    Parameters
    ----------
    D : float
        Diameter of the section
    B : float
        Width of the section
    load_coefficients : LoadCoefficients or dict
        Load coefficients for the section. If a dictionary is provided, it should contain
        the keys 'Cd', 'dCd', 'Cl', 'dCl', 'Cm', and 'dCm' with corresponding values.
        If a LoadCoefficients object is provided, it should have the same attributes.
    Returns
    -------
    dict
        A dictionary with the aerodynamic derivatives for the section.
    
    Notes
    -----
    The aerodynamic derivatives are calculated based on the load coefficients provided. 
    The load coefficients should be normalized with respect to both D (Cd) and B (Cl and Cm).
    ADs are normalized using B only.
    """

    if type(load_coefficients)==dict: 
        Cd = load_coefficients['Cd']
        dCd = load_coefficients['dCd']
        Cl = load_coefficients['Cl']
        dCl = load_coefficients['dCl']
        Cm = load_coefficients['Cm']
        dCm = load_coefficients['dCm']
    else:
        Cd, dCd = load_coefficients.Cd, load_coefficients.dCd
        Cl, dCl = load_coefficients.Cl, load_coefficients.dCl
        Cm, dCm = load_coefficients.Cm, load_coefficients.dCm

    ad_dict = dict()
    ad_dict['P1'], ad_dict['P2'], ad_dict['P3'] = lambda v: -2*Cd*D/B*v, lambda v: 0*v, lambda v: dCd*D/B*v**2
    ad_dict['P4'], ad_dict['P5'], ad_dict['P6'] = lambda v: 0*v, lambda v: (Cl-dCd*D/B)*v, lambda v: 0*v

    ad_dict['H1'], ad_dict['H2'], ad_dict['H3'] = lambda v: -(dCl+Cd*D/B)*v, lambda v: 0*v, lambda v: dCl*v**2
    ad_dict['H4'], ad_dict['H5'], ad_dict['H6'] = lambda v: 0*v, lambda v: -2*Cl*v, lambda v: 0*v

    ad_dict['A1'], ad_dict['A2'], ad_dict['A3'] = lambda v: -dCm*v, lambda v: 0*v, lambda v: dCm*v**2
    ad_dict['A4'], ad_dict['A5'], ad_dict['A6'] = lambda v: 0*v, lambda v: -2*Cm*v, lambda v: 0*v

    return ad_dict


def compute_aero_matrices(U, AD, B, elements, T_wind, phi, 
                          omega_reduced=None, print_progress=False, rho=1.225):
    """Compute the aerodynamic matrices for a given set of elements.
    
    Parameters
    ----------
    U : float or function
        Wind speed or function to calculate wind speed at a given point.
    AD : ADs
        Aerodynamic derivatives for the section.
    B : float
        Width of the section.
    elements : list
        List of elements for which to compute the aerodynamic matrices.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    phi : np.ndarray
        Mode shapes for the elements.
    omega_reduced : np.ndarray, optional
        Reduced frequencies for which to compute the aerodynamic matrices.
    print_progress : bool, optional
        If True, print progress of the computation.
    rho : float, optional
        Air density, default is 1.225 kg/m^3.
        
    Returns
    -------
    Kae : np.ndarray
        Aerodynamic stiffness matrix.
    Cae : np.ndarray    
        Aerodynamic damping matrix.
    """
    
    if omega_reduced is None:
        omega_reduced = np.linspace(0.015, 2.0, 75)
    
    n_modes = phi.shape[1]

    Kae = np.zeros([n_modes, n_modes, len(omega_reduced)])
    Cae = np.zeros([n_modes, n_modes, len(omega_reduced)])
    
    for element_ix, element in enumerate(elements):
        
        if callable(U):
            U_el_glob = U(element.get_cog())
        else:
            U_el_glob = U*1
            
        U_el = normal_wind(T_wind, element.T0, U=U_el_glob)

        v = U_el/(B*omega_reduced)
            
        for k, v_k in enumerate(v):
            k_aero, c_aero = element_aero_mats(B, omega_reduced[k], 
                                               AD.evaluate_all(v_k), 
                                               element.L, T=element.T0, 
                                               phi=phi[element.global_dofs, :], rho=rho)

            Kae[:, :, k] = Kae[:, :, k] + k_aero
            Cae[:, :, k] = Cae[:, :, k] + c_aero
        
        if print_progress:
            pp(element_ix+1, len(elements), sym='=', postfix=' ESTABLISHING WIND EXCITATION')
            print('')
    
    Cae = interp1d(omega_reduced, Cae, kind='quadratic', fill_value='extrapolate', bounds_error=False) 
    Kae = interp1d(omega_reduced, Kae, kind='quadratic', fill_value='extrapolate', bounds_error=False)


    return Kae, Cae


def compute_aero_matrices_sets(U, AD, B, elements, T_wind, phi_dict, 
                          omega_reduced=None, omega=None, print_progress=False, sets=None):
    
    """Compute the aerodynamic matrices for a given set of elements.
    
    Parameters
    ----------
    U : float or function
        Wind speed or function to calculate wind speed at a given point.
    AD : dict
        Dictionary of aerodynamic derivatives for the section.
    B : float
        Width of the section.
    elements : dict
        Dictionary of elements (BEEF elements) for which to compute the aerodynamic matrices.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    phi : np.ndarray
        Mode shapes for the elements.
    omega_reduced : np.ndarray, optional
        Reduced frequencies for which to compute the aerodynamic matrices.
    print_progress : bool, optional
        If True, print progress of the computation.
    sets : list, optional
        List of sets of elements for which to compute the aerodynamic matrices.

    Returns
    -------
    Kae : np.ndarray 
        Aerodynamic stiffness matrix.
    Cae : np.ndarray
        Aerodynamic damping matrix.
    """

    if sets is None:
         sets = elements.keys()

    if omega is None:
        return_as_function = True
    else:
        first_is_zero = omega[0]==0.0
        if first_is_zero:
            omega = omega[1:]
    
    if omega_reduced is None:
        omega_reduced = np.logspace(np.log10(0.01), np.log10(2), 100)     #standard values should be reasonable in most typical cases - change later!
    
    first_key = [str(key) for key in sets][0]
    n_modes = np.shape(phi_dict[first_key])[1]

    Kae = np.zeros([n_modes, n_modes, len(omega_reduced)])
    Cae = np.zeros([n_modes, n_modes, len(omega_reduced)])
    
    for set_name in sets:
        B_set = B[set_name]
        AD_set = AD[set_name]
        phi = phi_dict[set_name]
        elements_set = elements[set_name]

        for element_ix, element in enumerate(elements_set):
            T_el = element.T0
            U_el = normal_wind(T_wind, T_el, U=U)
            v = U_el/(B_set*omega_reduced)
            
            dof_range = np.hstack([element.nodes[0].global_dofs, element.nodes[1].global_dofs])

            for k, v_k in enumerate(v):
                k_aero, c_aero = element_aero_mats(B_set, omega_reduced[k], AD_set.evaluate_all(v_k), element.L, T=T_el, phi=phi[dof_range, :])
                Kae[:, :, k] += k_aero
                Cae[:, :, k] += c_aero
            
            if print_progress:
                pp(element_ix+1, len(elements_set), sym='>', postfix=f' finished with set "{set_name}".')

        if print_progress:
            print('')

        Cae = interp1d(omega_reduced, Cae, kind='quadratic',fill_value='extrapolate') 
        Kae = interp1d(omega_reduced, Kae, kind='quadratic', fill_value='extrapolate')

        if return_as_function:
            return Kae, Cae
        else:
            Cae = Cae(omega)
            Kae = Kae(omega)
            
            if first_is_zero:
                Cae = np.insert(Cae, 0, Cae[:,:,0]*0, axis=2)
                Kae = np.insert(Kae, 0, Kae[:,:,0]*0, axis=2)

            return Kae, Cae


def f_rf_fun(a, d, v):
    N = len(a) 
    f = np.array(a[0])*0j
    
    for l in range(0, 3):
        f = f + a[l] * (1j/v)**l

    for l in range(0, N-3):
        f = f + a[l+2]*(1j/v) / ((1j/v + d[l]))
    
    f = f*v**2    

    return f


def rf_ads(a, d):
    """ 
    Function to compute the aerodynamic derivatives for a given set of reduced frequencies using rational functions.

    Parameters
    ----------
    a : list
        List of coefficients for the rational function, assumed in the order of 
        P1, P2, P3, P4, P5, P6, H1, H2, H3, H4, H5, H6, A1, A2, A3, A4, A5, A6.
    d : list
        List of poles for the rational function, assumed in the order of 
        P1, P2, P3, P4, P5, P6, H1, H2, H3, H4, H5, H6, A1, A2, A3, A4, A5, A6.

    Returns
    -------
    dict
        A dictionary with the aerodynamic derivatives for the section.

    Notes
    -----
    The aerodynamic derivatives are calculated based on the coefficients and poles provided.
    The coefficients should be normalized with respect to both D (Cd) and B (Cl and Cm).
    
    """
    # B assumed to be implicitly included in RF factors
    ad_dict = dict()
    ad_keys = ['P1', 'P2', 'P3', 'P4', 'P5', 'P6', 
               'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 
               'A1', 'A2', 'A3', 'A4', 'A5', 'A6']
    
    imag_component_ad = ['P1', 'P2', 'P5', 'H1', 'H2', 'H5', 'A1', 'A2', 'A5']
    
    position_dict = {'P1': [0,0], 'P2': [0,2], 'P3': [0,2], 'P4': [0,0], 'P5': [0,1], 'P6': [0,1],
                 'H1': [1,1], 'H2': [1,2], 'H3': [1,2], 'H4': [1,1], 'H5': [1,0], 'H6': [1,0],
                 'A1': [2,1], 'A2': [2,2], 'A3': [2,2], 'A4': [2,1], 'A5': [2,0], 'A6': [2,0]}
    
    for key in ad_keys:
        row = position_dict[key][0]
        col = position_dict[key][1]
        a_key = [ai[row, col] for ai in a]

        if key in imag_component_ad:
            ad_dict[key] = lambda v, a=a_key: np.imag(f_rf_fun_legacy(a, d, v))
        else:
            ad_dict[key] = lambda v, a=a_key: np.real(f_rf_fun_legacy(a, d, v))

    return ad_dict


def distribute_to_dict(prefix, array, count_start=1):
    """ 
    Distributes a list of arrays to a dictionary with keys based on the prefix and index.

    Parameters
    ----------
    prefix : str
        Prefix for the keys in the dictionary.
    array : list
        List of arrays to be distributed.
    count_start : int, optional
        Starting index for the keys in the dictionary. Default is 1.

    Returns
    -------
    dict
        A dictionary with keys based on the prefix and index, and values from the input array.

    Notes
    -----
    The keys in the dictionary are formed by concatenating the prefix with the index of the array.

    Docstring is generated by GitHub Copilot.
    
    """
    array_dict = dict()
    for ix,array_i in enumerate(array):
        key = prefix + str(ix+count_start)
        array_dict[key] = array_i
        
    return array_dict
        

def distribute_multi_to_dict(prefixes, arrays):
    """
    Distributes a list of arrays to a dictionary with keys based on the prefix and index.

    Parameters
    ----------
    prefixes : list
        List of prefixes for the keys in the dictionary.
    arrays : list
        List of arrays to be distributed.

    Returns
    -------
    dict
        A dictionary with keys based on the prefix and index, and values from the input array.

    Notes
    -----
    The keys in the dictionary are formed by concatenating the prefix with the index of the array.
    Docstring is generated by GitHub Copilot."""
    array_dict = dict()
    
    for prefix_ix, prefix in enumerate(prefixes):
        for ix, array_i in enumerate(arrays[prefix_ix]):
            key = prefix + str(ix+1)
            array_dict[key] = array_i
            
    return array_dict


def unwrap_rf_parameters(parameters):
    """
    Unwraps the parameters from a dictionary into two lists: one for the coefficients and one for the poles.

    Parameters
    ----------
    parameters : dict
        Dictionary containing the parameters with keys starting with 'a' for coefficients and 'd' for poles.
    
    Returns
    -------
    tuple
        tuple containing two lists: one for the coefficients and one for the poles.
    
    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """

    keys = list(parameters.keys())
    a_ixs = np.where([word.startswith('a') for word in keys])[0]
    d_ixs  = np.where([word.startswith('d') for word in keys])[0]
    a_nums = np.array([int(string.split('a')[1]) for string in np.array(keys)[a_ixs]])
    d_nums = np.array([int(string.split('d')[1]) for string in np.array(keys)[d_ixs]])
    
    a = [np.zeros([3,3])]*(max(a_nums))
    d = [0]*(max(d_nums))
    
    for a_num in a_nums:
        a[a_num-1] = np.array(parameters['a%i' %a_num])
    
    for d_num in d_nums:
        d[d_num-1] = parameters['d%i' %d_num]
    
    d = np.array(d)
    return a,d


def normal_wind(T_g2wi, T_g2el, U=1.0):
    """
    Computes the wind speed in the local coordinates of the element.

    Parameters
    ----------
    T_g2wi : np.ndarray
        Transformation matrix from global to wind coordinates.
    T_g2el : np.ndarray 
        Transformation matrix from global to element coordinates.
    U : float, optional
        Wind speed in the global coordinates. Default is 1.0.
        
    Returns
    -------
    np.ndarray
        Wind speed in the local coordinates of the element.
    """

    T_wi2el = T_g2el @ T_g2wi.T
    e_wind_local = (T_wi2el @ np.array([1, 0, 0])[np.newaxis,:].T).flatten()
    
    Un = U * np.sqrt(e_wind_local[1]**2+e_wind_local[2]**2)
    return Un


def el_mat_generic(Ayy,Ayz,Ayt,Azy,Azz,Azt,Aty,Atz,Att,L):
    mat = np.zeros([12,12])

    mat[0:6, 0:6] = np.array([
        [0,         0,          0,          0,          0,              0           ],
        [0,         156*Ayy,    156*Ayz,    147*Ayt,    -22*L*Ayz,      22*L*Ayy    ],
        [0,         156*Azy,    156*Azz,    147*Azt,    -22*L*Azz,      22*L*Azy    ],
        [0,         147*Aty,    147*Atz,    140*Att,     -21*L*Atz,      21*L*Aty   ],
        [0,         -22*L*Azy,  -22*L*Azz,  -21*L*Azt,  4*L**2*Azz,     -4*L**2*Azy ],
        [0,         22*L*Ayy,   22*L*Ayz,   21*L*Ayt,   -4*L**2*Ayz,    4*L**2*Ayy  ],
    ])

    mat[0:6, 6:12] = np.array([
        [0,         0,          0,          0,          0,              0            ],
        [0,         54*Ayy,    54*Ayz,      63*Ayt,     13*L*Ayz,       -13*L*Ayy    ],
        [0,         54*Azy,    54*Azz,      63*Azt,     13*L*Azz,       -13*L*Azy    ],
        [0,         63*Aty,    63*Atz,      70*Att,     14*L*Atz,       -14*L*Aty    ],
        [0,         -13*L*Azy,  -13*L*Azz,  -14*L*Azt,  -3*L**2*Azz,     3*L**2*Azy  ],
        [0,         13*L*Ayy,   13*L*Ayz,   14*L*Ayt,   3*L**2*Ayz,     -3*L**2*Ayy  ],
    ])

    mat[6:12, 0:6] = np.array([
        [0,         0,          0,          0,          0,              0            ],
        [0,         54*Ayy,    54*Ayz,      63*Ayt,     -13*L*Ayz,       13*L*Ayy    ],
        [0,         54*Azy,    54*Azz,      63*Azt,     -13*L*Azz,       13*L*Azy    ],
        [0,         63*Aty,    63*Atz,      70*Att,     -14*L*Atz,       14*L*Aty    ],
        [0,         13*L*Azy,  13*L*Azz,    14*L*Azt,   -3*L**2*Azz,     3*L**2*Azy  ],
        [0,         -13*L*Ayy, -13*L*Ayz,   -14*L*Ayt,   3*L**2*Ayz,     -3*L**2*Ayy ],
    ])

    mat[6:12,6:12] = np.array([
        [0,         0,          0,          0,          0,              0               ],
        [0,         156*Ayy,    156*Ayz,    147*Ayt,    22*L*Ayz,      -22*L*Ayy        ],
        [0,         156*Azy,    156*Azz,    147*Azt,    22*L*Azz,      -22*L*Azy        ],
        [0,         147*Aty,    147*Atz,    140*Att,    21*L*Atz,      -21*L*Aty        ],
        [0,         22*L*Azy,   22*L*Azz,   21*L*Azt,    4*L**2*Azz,   -4*L**2*Azy      ],
        [0,         -22*L*Ayy,   -22*L*Ayz,   -21*L*Ayt,   -4*L**2*Ayz,    4*L**2*Ayy   ],
    ])

    return mat

def element_aero_mats(B, omega, ad_dict, L, T=None, phi=None, rho=1.225):
    # Called for selected reduced velocity, specified by omega value (implicitly mean wind).
    # Corresponding values of P,H and A are used for given mean wind velocity.

    # Stiffness
    Ayy = 1/2*rho*B**2*omega**2*ad_dict['P4']
    Ayz = 1/2*rho*B**2*omega**2*ad_dict['P6']
    Ayt = -1/2*rho*B**2*omega**2*B*ad_dict['P3']

    Azy = 1/2*rho*B**2*omega**2*ad_dict['H6']
    Azz = 1/2*rho*B**2*omega**2*ad_dict['H4']
    Azt = -1/2*rho*B**2*omega**2*B*ad_dict['H3']

    Aty = -1/2*rho*B**2*omega**2*B*ad_dict['A6']
    Atz = -1/2*rho*B**2*omega**2*B*ad_dict['A4']
    Att = 1/2*rho*B**2*omega**2*B**2*ad_dict['A3']

    k_aero = L/420 * el_mat_generic(Ayy,Ayz,Ayt,Azy,Azz,Azt,Aty,Atz,Att,L)

    
    # Damping
    Ayy = 1/2*rho*B**2*omega*ad_dict['P1']
    Ayz = 1/2*rho*B**2*omega*ad_dict['P5']
    Ayt = -1/2*rho*B**2*omega*B*ad_dict['P2']

    Azy = 1/2*rho*B**2*omega*ad_dict['H5']
    Azz = 1/2*rho*B**2*omega*ad_dict['H1']    
    Azt = -1/2*rho*B**2*omega*B*ad_dict['H2']
    
    Aty = -1/2*rho*B**2*omega*B*ad_dict['A5']
    Atz = -1/2*rho*B**2*omega*B*ad_dict['A1']
    Att = 1/2*rho*B**2*omega*B**2*ad_dict['A2']

    c_aero = L/420 * el_mat_generic(Ayy,Ayz,Ayt,Azy,Azz,Azt,Aty,Atz,Att,L)

    if (T is None and phi is None)!=True:
        if T is not None:     #if no transformation matrix is given, a local matrix is output
            if np.shape(T)[0]==6:
                T = np.kron(np.eye(2), T)   #two times 6dof matrix, block diagonal
            if np.shape(T)[0]==3:
                T = np.kron(np.eye(4), T)   #four times 3dof matrix, block diagonal
            elif np.shape(T)[0]!=12:
                raise ValueError('Wrong size of T (should be 3x3, 6x6 or 12x12')
        else:
            T = np.eye(12)
        
        if phi is not None:
            T = T @ phi
            
        k_aero = T.T @ k_aero @ T
        c_aero = T.T @ c_aero @ T
    
    return k_aero, c_aero


# Spectra
def kaimal_auto(omega, Lx, A, sigma, V):
    """
    Kaimal auto-spectral density function.

    Parameters
    ----------
    omega : float

        Angular frequency.
    Lx : float
        Length scale.
    A : float
        Amplitude.
    sigma : float
        Standard deviation of the turbulence.
    V : float   
        Mean wind speed.

    Returns
    -------
    S : float
        Auto-spectral density function value.
    """

    f = omega/(2*np.pi)
    fhat = f*Lx/V
    S = (sigma**2*(A*fhat)/(1+(1.5*A*fhat))**(5/3))/f

    return S/(2*np.pi)

def von_karman_auto(omega, Lx, sigma, V):
    """
    Von Karman auto-spectral density function.
    
    Parameters
    ----------
    omega : float
        Angular frequency.
    Lx : float
        Length scale.
    sigma : float
        Standard deviation of the turbulence.
    V : float
        Mean wind speed.
    Returns
    -------
    S : float
        Auto-spectral density function value.
    """
   
    A1 = [
        0.0,
        0.0,
        755.2,
        ]
    
    A2 = [
        70.8,
        0.0,
        283.2,
        ]
    
    rr = [
        5/6,
        11/6,
        11/6,
        ]
    
    f = omega/(2*np.pi)
    fhat = f*Lx/V
    S = (sigma**2*( (4*fhat)*(1+A1*fhat**2)  )/ (1+A2*fhat**2)**(rr))/f
    
    return S/(2*np.pi)

def generic_kaimal_matrix(omega, nodes, T_wind, A, sigma, C, Lx, U, spectrum_type='kaimal'):
    """
    Computes the cross-spectral density matrix for a given set of nodes.
    
    Parameters
    ----------
    omega : float

        Angular frequency.
    nodes : list
        List of BEEF nodes for which to compute the cross-spectral density matrix.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    A : float or np.ndarray
        Parameter(s) A in Kaimal auto-spectral density.
    sigma : float or np.ndarray
        Standard deviation of the turbulence (either single or all components).
    C : float or np.ndarray  
        Coefficient(s) for the cross-spectral density matrix.
    LX : float or np.ndarray
        Length scale(s) for the cross-spectral density matrix.
    U : function
        Function (of x, y and z) to calculate mean wind speed at a given point.
    spectrum_type : str, optional
        Type of spectrum to use. Default is 'kaimal'. Other options are 'vonKarman' or 'Karman'.

    """

    # Adopted from MATLAB version. `nodes` is list with beef-nodes.        
    V = np.zeros(len(nodes))      # Initialize vector with mean wind in all nodes
    Su = np.zeros([len(nodes), len(nodes)])     # One-point spectra for u component in all nodes
    Sv = np.zeros([len(nodes), len(nodes)])     # One-point spectra for v component in all nodes
    Sw = np.zeros([len(nodes), len(nodes)])     # One-point spectra for w component in all nodes
    xyz = np.zeros([len(nodes), 3])  # Nodes in wind coordinate system

    for node_ix, node in enumerate(nodes):
        xyz[node_ix,:] = (T_wind @ node.coordinates).T #Transform node coordinates to the wind coordinate system
        V[node_ix] = U(node.coordinates) # Mean wind velocity in the nodes

        if 'karman' in spectrum_type.lower():
            Su[node_ix,:], Sv[node_ix,:], Sw[node_ix,:] = von_karman_auto(omega, Lx, sigma, V[node_ix]) 
        elif spectrum_type.lower() == 'kaimal':
            Su[node_ix,:], Sv[node_ix,:], Sw[node_ix,:] = kaimal_auto(omega, Lx, A, sigma, V[node_ix]) # One point spectra for u component in all nodes
        else:
            raise ValueError('spectrum_type must either be defined as "vonKarman"/"Karman" or "Kaimal"')

    x = xyz[:, 0]
    y = xyz[:, 1]
    z = xyz[:, 2]
    
    dxdx = x[np.newaxis,:] - x[np.newaxis,:].T # Matrix with all distances between nodes in x direction
    dydy = y[np.newaxis,:] - y[np.newaxis,:].T # Matrix with all distances between nodes in y direction
    dzdz = z[np.newaxis,:] - z[np.newaxis,:].T # Matrix with all distances between nodes in z direction
    
    invV = 2/(V[np.newaxis,:]+V[np.newaxis,:].T) # Inverse mean wind velocity for all combination of nodes

    Suu = np.sqrt(Su)*np.sqrt(Su).T*np.exp(
            -invV*omega/(2*np.pi)*np.sqrt(
                (C[0,0]*dxdx)**2 + (C[1,0]*dydy)**2 + (C[2,0]*dzdz)**2)
        )
    
    Svv = np.sqrt(Sv)*np.sqrt(Sv).T*np.exp(
            -invV*omega/(2*np.pi)*np.sqrt(         
                (C[0,1]*dxdx)**2 + (C[1,1]*dydy)**2 + (C[2,1]*dzdz)**2)
        )
    
    Sww = np.sqrt(Sw)*np.sqrt(Sw).T*np.exp(
            -invV*omega/(2*np.pi)*np.sqrt( 
                (C[0,2]*dxdx)**2 + (C[1,2]*dydy)**2 + (C[2,2]*dzdz)**2)
        )
    
    SvSv = np.zeros([3*len(nodes), 3*len(nodes)]) # Cross sectral density matrix containing all the turbulence components
    SvSv[0::3, 0::3] = Suu
    SvSv[1::3, 1::3] = Svv
    SvSv[2::3, 2::3] = Sww
    
    return SvSv


def loadmatrix_fe(V, load_coefficients, rho, B, D, admittance=None):
    """
    Computes the aerodynamic load matrix for a given set of parameters.
    
    Parameters
    ----------
    V : float
        Wind speed.
    load_coefficients : dict
        Dictionary containing the aerodynamic coefficients.
    rho : float
        Air density.
    B : float
        Width of the section.
    D : float
        Height of the section.
    admittance : function, optional
        Function to compute the admittance function. Default is None.
        
    Returns
    -------
    BqBq : function
        Function to compute the aerodynamic load matrix for a given frequency.
    """

    if admittance is None :
        admittance = lambda omega_k: np.ones( (4,3) )

    Cd = load_coefficients['Cd']
    dCd = load_coefficients['dCd']
    Cl = load_coefficients['Cl']
    dCl = load_coefficients['dCl']
    Cm = load_coefficients['Cm']
    dCm = load_coefficients['dCm']
    
    # Equation 7 from Oiseth, 2010
    BqBq = lambda omega_k: 1/2*rho*V*B*admittance(omega_k*B/V/2/np.pi)*np.array([[0,            0,          0],
                                                                                [0,     2*D/B*Cd,          (D/B*dCd-Cl)],
                                                                                [0,         2*Cl,          (dCl+D/B*Cd)],
                                                                                [0,      -2*B*Cm,          -B*dCm]])
                                          
    return BqBq

def loadmatrix_fe_static(V, load_coefficients, rho, B, D ):
    """
    Computes the aerodynamic load matrix for a given set of parameters in static conditions.
    
    Parameters
    ----------
    V : float
        Wind speed.
    load_coefficients : dict
        Dictionary containing the aerodynamic coefficients.
    rho : float
        Air density.
    B : float
        Width of the section.
    D : float
        Height of the section.
    
    Returns
    -------
    BqBq : np.ndarray
        Aerodynamic load matrix.
    """
    
    Cd = load_coefficients['Cd']
    Cl = load_coefficients['Cl']
    Cm = load_coefficients['Cm']
        
    BqBq = 1/2*rho*V**2*B*np.array([[ 0 ],
                                    [ D/B*Cd ],
                                    [ Cl  ],
                                    [ -B*Cm ]])
    return BqBq

def loadvector(T_el, Bq, T_wind, L, static = False):
    """
    Computes load vectors for a given set of parameters.
    
    Parameters
    ----------
    T_el : np.ndarray
        Transformation matrix from global to element coordinates.
    Bq : np.ndarray
        Aerodynamic load matrix.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    L : float
        Length of the element.
    static : bool, optional
        If True, compute the load vector for static conditions. Default is False.

    Returns
    -------
    R1 : np.ndarray
        Load vector for element node 1.
    R2 : np.ndarray
        Load vector for element node 2.

    """

    G = np.zeros([12,4])
    G[0,0] = L/2
    G[1,1] = L/2
    G[2,2] = L/2
    G[3,3] = L/2
    G[6,0] = L/2
    G[7,1] = L/2
    G[8,2] = L/2
    G[9,3] = L/2
    G[4,2] = -L**2/12
    G[5,1] = L**2/12
    G[10,2] = L**2/12
    G[11,1] = -L**2/12
 
    # Transform from wind coordinates to local element coordinates 
    
    T = T_el @ T_wind.T   
    
    T_full = blkdiag(T_el, 4)     # Block diagonal - repeated 4 times to transform both trans and rot DOFs at each node (2+2)
    
    # T_full.T transforms L-->G
    if static is False:
        R =  T_full.T @ G @ Bq @ T 
    else:
        R =  T_full.T @ G @ Bq 
        
    R1 =  R[0:6]         # Element node 1
    R2 =  R[6:12]        # Element node 2


    return R1, R2


def windaction(omega, S, load_coefficients, elements, T_wind, 
               phi, B, D, U, omega_reduced=None, rho=1.225, print_progress=True,
               section_lookup=None, nodes=None, admittance=None):
    """
    Computes the wind cross-spectral density matrix on a set of elements using the specified parameters.
    
    Parameters
    ----------
    omega : np.ndarray
        Angular frequency.
    S : function
        Function to compute the cross-spectral density matrix.
    load_coefficients : dict
        Dictionary containing the aerodynamic coefficients.
    elements : list
        List of elements for which to compute the wind cross-spectral density matrix.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    phi : np.ndarray
        Transformation matrix from global to element coordinates.
    B : float  
        Width of the section.
    D : float
        Height of the section.
    U : function
        Function to calculate mean wind speed at a given point.
    omega_reduced : np.ndarray, optional    
        Reduced angular frequency. Default is None.
    rho : float, optional
        Air density. Default is 1.225.
    print_progress : bool, optional
        If True, print progress. Default is True.
    section_lookup : dict, optional
        Dictionary to look up sections for each element. Default is None.
    nodes : list, optional
        List of BEEF nodes for which to compute the wind cross-spectral density matrix. Default is None.
    admittance : function, optional
        Function to compute the admittance function. Default is None.
    
    Returns
    ------- 
    genSqSq_reduced : np.ndarray
        Cross-spectral density matrix for the wind action.

    """
    
    if nodes is None:
        nodes = list(set([a for b in [el.nodes for el in elements] for a in b]))

    n_dofs = 6
    
    # Ensure that first omega value is not 0 when using logspace omega axis
    if omega_reduced is None:
        if np.min(omega) == 0:
            omega_sorted = np.sort(omega)
            omega_start = omega_sorted[1]
        else:
            omega_start = np.min(omega)    
        
        omega_reduced = np.logspace(np.log10(omega_start), np.log10(np.max(omega)), num=50) # A log frequency axis that is used to obtain the cross-spectral density matrix
        
    genSqSq_reduced = np.zeros([phi.shape[1], phi.shape[1], len(omega_reduced)]) # Initialize the cross-spectral density matrix
    
    # Establish RG matrix (common for all freqs)
    

    if section_lookup is None:
        lc_fun = lambda el: load_coefficients
        B_fun = lambda el: B
        D_fun = lambda el: D
        admittance_fun = lambda el: admittance
    else:
        def get_sec(el):
            for key in section_lookup:
                if el in section_lookup[key]:    
                    return key 

        lc_fun = lambda el: load_coefficients[get_sec(el)]
        B_fun = lambda el: B[get_sec(el)]
        D_fun = lambda el: D[get_sec(el)]
    
    if admittance is None: # omit the frequency loop if ADmittance is not included - faster !
        RG = np.zeros([len(nodes)*n_dofs, 3])
        for el in elements:            
            node1_dofs = el.nodes[0].global_dofs
            node2_dofs = el.nodes[1].global_dofs

            mean_wind = U(el.get_cog())         
            Vn = normal_wind(T_wind, el.T0)*mean_wind        # Find the normal wind
            BqBq = loadmatrix_fe(Vn, lc_fun(el), rho, B_fun(el), D_fun(el))
            R1, R2 = loadvector(el.T0, BqBq(1), T_wind, el.L) # Obtain the load vector for each element

            RG[node1_dofs, :] = RG[node1_dofs, :] + R1   # Add the contribution from the element (end 1) to the system
            RG[node2_dofs, :] = RG[node2_dofs, :] + R2   # Add the contribution from the element (end 2) to the system

            # Make block matrix
            RG_block = np.zeros([6*len(nodes), 3*len(nodes)])
            
            for node in nodes:
                ix = node.index
                n = np.r_[6*ix:6*ix+6]
                m = np.r_[3*ix:3*ix+3]
                RG_block[np.ix_(n,m)] = RG[n,:]     #verified with MATLAB version for beam example

            for k, omega_k in enumerate(omega_reduced): 
                if print_progress:
                    pp(k+1, len(omega_reduced), sym='=', postfix=' ESTABLISHING WIND EXCITATION')
                    print('')
                    
                phiT_RG_block = phi.T @ RG_block
                genSqSq_reduced[:, :, k] = phiT_RG_block @ S(omega_k) @ phiT_RG_block.T  # to modal coordinates

    else: # admittance is given - triple loop (the old way, slower)
        admittance_fun = lambda el: admittance[get_sec(el)]

        for k, omega_k in enumerate(omega_reduced): 
            if print_progress:
                pp(k+1, len(omega_reduced), sym='=', postfix=' ESTABLISHING WIND EXCITATION')
                print('')
        
            # Establish RG matrix 
            RG = np.zeros([len(nodes)*n_dofs, 3])
            
            for el in elements:            
                node1_dofs = el.nodes[0].global_dofs
                node2_dofs = el.nodes[1].global_dofs

                mean_wind = U(el.get_cog())         
                Vn = normal_wind(T_wind, el.T0)*mean_wind        # Find the normal wind
                BqBq = loadmatrix_fe(Vn, lc_fun(el), rho, B_fun(el), D_fun(el), admittance=admittance_fun(el))
                R1, R2 = loadvector(el.T0, BqBq(omega_k), T_wind, el.L) # Obtain the load vector for each element

                RG[node1_dofs, :] = RG[node1_dofs, :] + R1   # Add the contribution from the element (end 1) to the system
                RG[node2_dofs, :] = RG[node2_dofs, :] + R2   # Add the contribution from the element (end 2) to the system
                
                
            # Make block matrix
            RG_block = np.zeros([6*len(nodes), 3*len(nodes)])
            
            for node in nodes:
                ix = node.index
                n = np.r_[6*ix:6*ix+6]
                m = np.r_[3*ix:3*ix+3]
                RG_block[np.ix_(n,m)] = RG[n,:]     #verified with MATLAB version for beam example
                
            phiT_RG_block = phi.T @ RG_block
            genSqSq_reduced[:, :, k] = phiT_RG_block @ S(omega_k) @ phiT_RG_block.T  # to modal coordinates


    # Interpolate results to full frequency axis
    genSqSq = interp1d(omega_reduced, genSqSq_reduced, kind='quadratic', axis=2, fill_value=0, bounds_error=False)
    
    return genSqSq

def windaction_static(load_coefficients, elements, T_wind, 
               phi, B, D, U, rho=1.225, print_progress=True,
               section_lookup=None, nodes=None):
    
    """
    Computes the wind load vector for a set of elements using the specified parameters.
    
    Parameters
    ----------
    load_coefficients : dict
        Dictionary containing the aerodynamic coefficients.
    elements : list
        List of BEEF elements for which to compute the wind load vector.
    T_wind : np.ndarray
        Transformation matrix from global to wind coordinates.
    phi : np.ndarray
        Transformation matrix from global to element coordinates.
    B : float
        Width of the section.
    D : float
        Height of the section.
    U : function
        Function to calculate mean wind speed at a given point.
    rho : float, optional
        Air density. Default is 1.225.
    print_progress : bool, optional
        If True, print progress. Default is True.
    section_lookup : dict, optional
        Dictionary to look up sections for each element. Default is None.
    nodes : list, optional
        List of BEEF nodes for which to compute the wind load vector. Default is None.
    
    Returns
    -------
    genF : np.ndarray
        Wind load vector for the specified elements.
        """
    
    if nodes is None:
        nodes = list(set([a for b in [el.nodes for el in elements] for a in b]))

    n_dofs = 6
    
    if section_lookup is None:
        lc_fun = lambda el: load_coefficients
        B_fun = lambda el: B
        D_fun = lambda el: D
    else:
        def get_sec(el):
            for key in section_lookup:
                if el in section_lookup[key]:    
                    return key 

        lc_fun = lambda el: load_coefficients[get_sec(el)]
        B_fun = lambda el: B[get_sec(el)]
        D_fun = lambda el: D[get_sec(el)]
            
        # Establish RG matrix 
        RG = np.zeros([len(nodes)*n_dofs])
        
        for el in elements:            
            node1_dofs = el.nodes[0].global_dofs
            node2_dofs = el.nodes[1].global_dofs

            mean_wind = U(el.get_cog())         
            Vn = normal_wind(T_wind, el.T0)*mean_wind        # Find the normal wind
            BqBq = loadmatrix_fe_static(Vn, lc_fun(el), rho, B_fun(el), D_fun(el))
            R1, R2 = loadvector(el.T0, BqBq, T_wind, el.L, static=True) # Obtain the load vector for each element

            RG[node1_dofs] = RG[node1_dofs] + R1[:,0]   # Add the contribution from the element (end 1) to the system
            RG[node2_dofs] = RG[node2_dofs] + R2[:,0]    # Add the contribution from the element (end 2) to the system
            
        # Make block matrix
        RG_block = np.zeros([6*len(nodes)])
        
        for node in nodes:
            ix = node.index
            n = np.r_[6*ix:6*ix+6]
            RG_block[np.ix_(n)] = RG[n]
            
        genF = phi.T @ RG_block 
    
    return genF

def K_from_ad(ad, V, w, B, rho):
    """
    Compute the aerodynamic stiffness matrix K from aerodynamic derivatives.

    Parameters
    ----------
    ad : dict
        Dictionary containing aerodynamic derivative functions. Expected keys are:
        'P4', 'P6', 'P3', 'H6', 'H4', 'H3', 'A6', 'A4', 'A3'. Each value should be a callable
        that takes a single argument (the reduced velocity `v`).
    V : float
        Wind velocity.
    w : float
        Circular frequency (rad/s). If zero, returns a zero matrix.
    B : float
        Reference length (e.g., bridge deck width).
    rho : float
        Air density.

    Returns
    -------
    k : ndarray of shape (3, 3)
        Aerodynamic stiffness matrix.

    Notes
    -----
    If `w` is zero, the function returns a 3x3 zero matrix.
    Otherwise, the matrix is computed using the provided aerodynamic derivatives evaluated at the reduced velocity.
    """
    
    if w==0:
        k = np.zeros([3,3])
    else:
        v = V / (B*w) # reduced velocity  
        
        k = (0.5*rho*B**2*w**2 * 
                np.vstack([[ad['P4'](v), ad['P6'](v), -B*ad['P3'](v)],
                           [ad['H6'](v), ad['H4'](v), -B*ad['H3'](v)],
                           [-B*ad['A6'](v), -B*ad['A4'](v), B**2*ad['A3'](v)]]))
        
        
    return k


def C_from_ad(ad, V, w, B, rho):
    """
    Compute the aerodynamic damping matrix K from aerodynamic derivatives.

    Parameters
    ----------
    ad : dict
        Dictionary containing aerodynamic derivative functions. Expected keys are: 
        'P1', 'P5', 'P2', 'H5', 'H1', 'H2', 'A5', 'A1', 'A2'. Each value should be a callable
        that takes a single argument (the reduced velocity `v`).
    V : float
        Wind velocity.
    w : float
        Circular frequency (rad/s). If zero, returns a zero matrix.
    B : float
        Reference length (e.g., bridge deck width).
    rho : float
        Air density.

    Returns
    -------
    k : ndarray of shape (3, 3)
        Aerodynamic stiffness matrix.
        
    Notes
    -----
    If `w` is zero, the function returns a 3x3 zero matrix.
    Otherwise, the matrix is computed using the provided aerodynamic derivatives evaluated at the reduced velocity.
    """
    if w==0:
        c = np.zeros([3,3])
    else:
        v = V / (B*w) # reduced velocity    
    
        c = (0.5*rho*B**2*w * 
            np.vstack([[ad['P1'](v), ad['P5'](v), -B*ad['P2'](v)],
                      [ad['H5'](v), ad['H1'](v), -B*ad['H2'](v)],
                      [-B*ad['A5'](v), -B*ad['A1'](v), B**2*ad['A2'](v)]]))  
    
    return c


def phi_aero_sum(mat, phi, x):
    """
    Computes the aerodynamic matrix (damping or stiffness) in modal coordinates.

    Parameters
    ----------
    mat : ndarray
        Aerodynamic matrix in global coordinates.
    phi : ndarray
        Transformation matrix from global to modal coordinates.
    x : ndarray
        Vector of x-coordinates for integration.

    Returns
    -------
    mat : ndarray
        Aerodynamic matrix in modal coordinates.
    """

    n_modes = phi.shape[1]
    n_points = len(x)
    
    mat_int = np.zeros([n_modes, n_modes, n_points])
    
    for p in range(n_points):
        phi_point = phi[p*6+1:p*6+4, :]
        mat_int[:, :, p] = phi_point.T @ mat @ phi_point

    mat = np.trapz(mat_int, x=x, axis=2)

    return mat


def function_sum(fun, const, fun_factor=1):
    """
    Function to sum a function and a constant. The function is multiplied by a factor.
    
    Parameters
    ----------
    fun : callable
        Function to be summed with the constant.
    const : float
        Constant to be summed with the function.
    fun_factor : float, optional
        Factor to multiply the function. Default is 1.
    
    Returns
    -------
    fsum : callable
        Function that computes the sum of the function and the constant.
        
    """
    def fsum(x):
        if fun is None:
            return const
        else:
            return fun(x)*fun_factor + const
    
    return fsum


def get_aero_cont_adfun(ad_dict_fun, V, B, rho, phi, x):
    """
    Computes the aerodynamic matrices in modal coordinates.

    Parameters
    ----------
    ad_dict_fun : callable
        Function that returns the aerodynamic derivatives for a given reduced velocity.
    V : float
        Wind velocity.
    B : float
        Reference length (e.g., bridge deck width).
    rho : float
        Air density.
    phi : ndarray
        Transformation matrix from global to modal coordinates.
    x : ndarray
        Vector of x-coordinates for integration.

    Returns
    ------- 
    K : callable
        Function that computes the aerodynamic stiffness matrix in modal coordinates.
    C : callable
        Function that computes the aerodynamic damping matrix in modal coordinates.
    
    
    """
    def K(w):
        n_modes = phi.shape[1]
        n_points = len(x)
        
        mat_int = np.zeros([n_modes, n_modes, n_points])
        
        for p in range(n_points):
            phi_point = phi[p*6+1:p*6+4, :]
            kae = K_from_ad(ad_dict_fun(x[p]), V, w, B, rho)
            mat_int[:, :, p] = phi_point.T @ kae @ phi_point

        return np.trapz(mat_int, x=x, axis=2)
    
    
    def C(w):
        n_modes = phi.shape[1]
        n_points = len(x)
        
        mat_int = np.zeros([n_modes, n_modes, n_points])
        
        for p in range(n_points):
            phi_point = phi[p*6+1:p*6+4, :]
            kae = C_from_ad(ad_dict_fun(x[p]), V, w, B, rho)
            mat_int[:, :, p] = phi_point.T @ kae @ phi_point

        return np.trapz(mat_int, x=x, axis=2)
    
        
    return K, C


def get_aero_cont_addict(ad_dict, V, B, rho, phi, x):
    """ 
    Computes the aerodynamic matrices in modal coordinates.

    Parameters
    ----------
    ad_dict : dict
        Dictionary containing aerodynamic derivative functions. Expected keys are:
        'P4', 'P6', 'P3', 'H6', 'H4', 'H3', 'A6', 'A4', 'A3'. Each value should be a callable
        that takes a single argument (the reduced velocity `v`).
    V : float
        Wind velocity.
    B : float
        Reference length (e.g., bridge deck width).
    rho : float
        Air density.
    phi : ndarray
        Transformation matrix from global to modal coordinates.
    x : ndarray
        Vector of x-coordinates for integration.

    Returns
    -------
    K : callable
        Function that computes the aerodynamic stiffness matrix in modal coordinates.
    C : callable
        Function that computes the aerodynamic damping matrix in modal coordinates.
    
    
    """
    def K(w):
        kae = K_from_ad(ad_dict, V, w, B, rho)
        return phi_aero_sum(kae, phi, x)
    
    def C(w):
        cae = C_from_ad(ad_dict, V, w, B, rho)
        return phi_aero_sum(cae, phi, x)
        
    return K, C


def itflutter_cont(Ms, Cs, Ks, phi, x, ad_dict, B, V=0.0, rho=1.225, dV=1, 
              overshoot_factor=0.5, itmax={}, omega_ref=None,
              tol={}, print_progress=True, keep_all=False, track_by_psi=True):
    
    """
    Iterative flutter analysis for continuous systems using the iterative eigensolver.
    
    Parameters
    ----------
    Ms : ndarray
        Mass matrix for in-vacuum and dry sub-structure.
    Cs : ndarray
        Damping matrix for in-vacuum and dry sub-structure.
    Ks : ndarray
        Stiffness matrix for in-vacuum and dry sub-structure.
    phi : ndarray
        Transformation matrix from global to modal coordinates.
    x : ndarray
        Vector of x-coordinates for integration.
    ad_dict : dict or callable
        Dictionary containing aerodynamic derivative functions or a function that returns the aerodynamic derivatives.
        If a dictionary, it should contain the following keys: 'P4', 'P6', 'P3', 'H6', 'H4', 'H3', 'A6', 'A4', 'A3'.
        Each value should be a callable that takes a single argument (the reduced velocity `v`).
    B : float
        Reference length (e.g., bridge deck width).
    V : float, optional
        Initial wind velocity. Default is 0.0.
    rho : float, optional
        Air density. Default is 1.225.
    dV : float, optional
        Increment for wind velocity. Default is 1.
    overshoot_factor : float, optional
        Factor to reduce the increment for wind velocity in case of overshooting. Default is 0.5.
    itmax : dict, optional
        Dictionary containing maximum iterations for velocity and frequency. Default is {'V': 50, 'f': 15}.
    omega_ref : float, optional
        Reference angular frequency. If None, it is computed from the eigenvalues of the system.
    tol : dict, optional
        Dictionary containing tolerance for velocity and frequency. Default is {'V': 1e-3, 'f': 1e-4}.
    print_progress : bool, optional
        If True, print progress. Default is True.
    keep_all : bool, optional
        If True, keep all results. Default is False.
    track_by_psi : bool, optional
        If True, track the eigenvectors by their similarity. Default is True.

    Returns
    -------
    res : dict
        Dictionary containing the results of the flutter analysis. The keys are:
    'V' : list of wind velocities at which flutter was detected.
    'lambd' : list of eigenvalues at which flutter was detected.
    'critical_mode' : list of indices of the critical mode at each velocity.
    'critical_psi' : list of eigenvectors corresponding to the critical mode at each velocity.
    """
    
    if callable(ad_dict):
        get_aero = get_aero_cont_adfun
    else:
        get_aero = get_aero_cont_addict

    itmax_ = {'V':50, 'f': 15}
    itmax_.update(**itmax)
    itmax = dict(itmax_)

    tol_ = {'V': 1e-3, 'f': 1e-4}
    tol_.update(**tol)
    tol = tol_
    
    res = dict()
    res['V'] = []
    res['lambd'] = []
    res['critical_mode'] = []
    res['critical_psi'] = []
    
    converged = False
    psi_prev = None
    
    if omega_ref is None:
        A = statespace(Ks, Cs, Ms)
        lambd_ref, psi = np.linalg.eig(A)  
        omega_initial = np.sort(np.abs(np.imag(lambd_ref)))[::2]
        omega_ref = omega_initial[0]
        
    for it_vel in range(itmax['V']):
        Kae, Cae = get_aero(ad_dict, V, B, rho, phi, x)
        getK = function_sum(Kae, Ks, fun_factor=-1)
        getC = function_sum(Cae, Cs, fun_factor=-1)
        getM = function_sum(None, Ms, fun_factor=-1)

        lambd, psi, not_converged = iteig(getK, getC, getM, tol=tol['f'], 
                                          keep_full=True, mac_min=0.0, itmax=itmax['f'])

        if len(not_converged)>0:
            lambd[not_converged] = -np.inf + 0j
            if print_progress:
                if len(not_converged)<10:
                    nc_modes = 'index '+ ', '.join([str(i) for i in not_converged])
                else:
                    nc_modes = '>10'
                print(f'** Non-converged modes ({nc_modes}) from iterative eigensolution disregarded! **')
                
        if it_vel!=0 and track_by_psi:
            ixs, __, __, __ = restructure_as_ref(psi_prev, psi)
            
            psi = psi[:, ixs]
            lambd = lambd[ixs] 
        
        psi_prev = psi*1
            
        critical_mode = np.argmax(np.real(lambd))
        real_lambd = np.max(np.real(lambd)) 
        critical_omega = np.abs(np.imag(lambd[critical_mode]))
        
        if keep_all or real_lambd<=0:
            res['critical_mode'].append(critical_mode)
            res['lambd'].append(lambd)
            res['V'].append(V)
            res['critical_psi'].append(psi[:,critical_mode])
        
        if dV < tol['V'] and real_lambd<=0:
            converged = True
            if print_progress:
                print(conv_text)
                print(f'Flutter estimated to occur at V = {V:.2f} m/s ({critical_omega:.2f} rad/s) ==>  v = {V/(B*critical_omega):.2f})\n')
            
            break
        elif real_lambd<0:
            if print_progress:
                print(f'Increasing velocity V = {V:.2f} --> {V+dV:.2f}.')
            V = V + dV
        else:
            if print_progress:
                print(f'Overshot. Reducing velocity V = {V:.2f} --> {V-dV/2:.2f}. Reducing step size dV = {dV:.2f} --> {dV/2:.2f}')
                
            dV = overshoot_factor*dV  # adjusting the velocity increment, and step backwards   
            V = V - dV

    if not converged and print_progress:
        print('Not able to converge within specified maximum iterations for specified tolerance criteria.')
    
    res = {key: np.array(res[key]) for key in ['critical_mode', 'critical_psi', 'V', 'lambd']}
    
    return res



def itflutter_cont_naive(Ms, Cs, Ks, phi, x, ad_dict, B, V=0.0, rho=1.225, dV=1, 
              overshoot_factor=0.5, itmax={}, tol={}, print_progress=True):
    """
    Iterative flutter analysis for continuous systems using a naive iteration.
    
    Parameters
    ----------
    Ms : ndarray
        Mass matrix for in-vacuum and dry sub-structure.
    Cs : ndarray
        Damping matrix for in-vacuum and dry sub-structure.
    Ks : ndarray
        Stiffness matrix for in-vacuum and dry sub-structure.
    phi : ndarray
        Transformation matrix from global to modal coordinates.
    x : ndarray
        Vector of x-coordinates for integration.
    ad_dict : dict or callable
        Dictionary containing aerodynamic derivative functions or a function that returns the aerodynamic derivatives.
        If a dictionary, it should contain the following keys: 'P4', 'P6', 'P3', 'H6', 'H4', 'H3', 'A6', 'A4', 'A3'.
        Each value should be a callable that takes a single argument (the reduced velocity `v`).
    B : float
        Reference length (e.g., bridge deck width).
    V : float, optional 
        Initial wind velocity. Default is 0.0.
    rho : float, optional   
        Air density. Default is 1.225.
    dV : float, optional
        Increment for wind velocity. Default is 1.
    overshoot_factor : float, optional
        Factor to reduce the increment for wind velocity in case of overshooting. Default is 0.5.
    itmax : dict, optional
        Dictionary containing maximum iterations for velocity and frequency. Default is {'V': 50, 'f': 15}.
    tol : dict, optional
        Dictionary containing tolerance for velocity and frequency. Default is {'V': 1e-3, 'f': 1e-4}.
    print_progress : bool, optional
        If True, print progress. Default is True.

    Returns
    ------- 
    res : dict
        Dictionary containing the results of the flutter analysis. The keys are:
    'V' : list of wind velocities at which flutter was detected.
    'lambd' : list of eigenvalues at which flutter was detected.
    'critical_mode' : list of indices of the critical mode at each velocity.
    'critical_psi' : list of eigenvectors corresponding to the critical mode at each velocity.

    """
        
    if callable(ad_dict):
        get_aero = get_aero_cont_adfun
    else:
        get_aero = get_aero_cont_addict
    
    itmax_ = {'V':50, 'f': 15}
    itmax_.update(**itmax)
    itmax = itmax_
    
    tol_ = {'V': 1e-3, 'f': 1e-4}
    tol_.update(**tol)
    tol = tol_
    
    res = dict()
    res['V'] = []
    res['lambd'] = []
    res['critical_mode'] = []
    res['critical_psi'] = []
    
    converged = False
        
    for it_vel in range(itmax['V']):
        Kae, Cae = get_aero(ad_dict, V, B, rho, phi, x)
        getK = function_sum(Kae, Ks, fun_factor=-1)
        getC = function_sum(Cae, Cs, fun_factor=-1)
        getM = function_sum(None, Ms, fun_factor=-1)

        lambd, psi = iteig_naive(getK, getC, getM, tol=tol['f'], itmax=itmax['f'])
        
        complex_ix = np.imag(lambd) != 0
        
        critical_mode = np.argmax(np.real(lambd[complex_ix]))
        critical_mode = np.where(complex_ix)[0][critical_mode]
        
        real_lambd = np.max(np.real(lambd)) 
        critical_omega = np.abs(np.imag(lambd[critical_mode]))
        
        if real_lambd<=0:
            res['critical_mode'].append(critical_mode)
            res['lambd'].append(lambd)
            res['V'].append(V)
            res['critical_psi'].append(psi[:,critical_mode])
        
        if dV < tol['V'] and real_lambd<=0:
            
            converged = True
            if print_progress:

                print(conv_text)
                print(f'Flutter estimated to occur at V = {V:.2f} m/s ({critical_omega:.2f} rad/s) ==>  v = {V/(B*critical_omega):.2f})\n')
            
            break
        elif real_lambd<=0:
            if print_progress:
                print(f'Increasing velocity V = {V:.2f} --> {V+dV:.2f}.')
            V = V + dV
        else:
            if print_progress:
                print(f'Overshot. Reducing velocity V = {V:.2f} --> {V-dV/2:.2f}. Reducing step size dV = {dV:.2f} --> {dV/2:.2f}')
                
            dV = overshoot_factor*dV  # adjusting the velocity increment, and step backwards   
            V = V - dV

    if not converged and print_progress:
        print('Not able to converge within specified maximum iterations for specified tolerance criteria.')
    
    res = {key: np.array(res[key]) for key in ['critical_mode', 'critical_psi', 'V', 'lambd']}
    
    return res


from sys import path
from collections import OrderedDict

path.append('C:/Users/knutankv/git-repos/wawi/')   # this is an easy quick fix to enable importing wawi package in Abaqus environment
savefolder = 'C:/Temp/'

# NOTE: Only linear beam elements are supported in relevant sets!!

import wawi.ext.abq
import json
import numpy as np

#%% Get (O) database object
db = wawi.ext.abq.get_db('odb')

#%% Definitions
frequency_step = 'Step-1'
part = db.rootAssembly.instances['TRUSSPART']
step_obj = db.steps[frequency_step]

if 'ALL' not in part.elementSets:   #CREATE SET OF ALL ELEMENTS IN PART
    part.ElementSet('ALL', part.elements)

#%% Grab regions
region_full = part.elementSets['ALL']
region_hydro = db.rootAssembly.nodeSets['PALL']

#%% Get modal parameters
fn, m = wawi.ext.abq.get_modal_parameters(frequency_step)

#%% Get wind elements and mode shapes
node_matrix, element_matrix = wawi.ext.abq.get_element_matrices(region_full, obj=part)
node_labels = node_matrix[:,0]
nodes = wawi.ext.abq.create_list_of_nodes(part, node_labels)

# Export element definitions as json
el_data = dict(node_matrix=node_matrix.tolist(), element_matrix=element_matrix.tolist())

with open('element.json', 'w') as f:
    json.dump(el_data, f)

phi_full_disp = wawi.ext.abq.get_nodal_phi(step_obj, nodes, flatten_components=True, field_outputs=['UT', 'UR'])

#%% Get pontoon data
pontoon_sets = ['P1', 'P2', 'P3', 'P4', 'P5', 'P6', 'P7']
pontoon_types = ['p17', 'p26', 'p345', 'p345', 'p345', 'p26', 'p17']
rotations = np.array([13.944407004001892, 9.2962713360012632, 4.6481356680006325, 0.0,
                     -4.6481356680006334, -9.2962713360012632, -13.944407004001883])

node_matrix_pontoons = np.vstack([wawi.ext.abq.get_element_matrices(db.rootAssembly.nodeSets[pset], obj=part) for pset in pontoon_sets])
node_labels_pontoons = node_matrix_pontoons[:,0]

# Export pontoon.json
pontoon_data = OrderedDict()

for ix, node in enumerate(node_labels_pontoons):
    key = pontoon_sets[ix]
    pontoon_data[key] = dict(coordinates=node_matrix_pontoons[ix, 1:].tolist(),
                            node=node,
                            rotation=rotations[ix],
                            pontoon_type=pontoon_types[ix])


nodes_pontoons = wawi.ext.abq.create_list_of_nodes(part, node_labels_pontoons)
phi_h = wawi.ext.abq.get_nodal_phi(step_obj, nodes_pontoons, flatten_components=True, field_outputs=['UT', 'UR'])


with open('pontoon.json', 'w') as f:
    json.dump(pontoon_data, f)

## ------------------- EXPORT MODES ----------------
modal_data = dict(omega_n=(fn*2*np.pi).tolist(), m=m.tolist(), phi=dict(full=phi_full_disp.tolist(),
                                     hydro=phi_h.tolist()))

with open('modal.json', 'w') as f:
    json.dump(modal_data, f)


from .wave import jonswap
import numpy as np
from scipy.optimize import curve_fit

def fit_jonswap(S, w, sigma=[0.07, 0.09], initial_values=None):
    """
    Fit the JONSWAP spectrum to a given numeric array describing spectral density.

    Parameters
    ----------
    S : array_like
        Spectral density values.
    w : array_like
        Angular frequency values corresponding to `S`.
    sigma : list of float, optional
        Sigma parameters for the JONSWAP spectrum. Default is [0.07, 0.09].
    initial_values : dict, optional
        Dictionary of initial values for the parameters 'Hs', 'Tp', and 'gamma'.
        If not provided, default values are used.

    Returns
    -------
    out : dict
        Dictionary containing the fitted parameters:
        - 'Hs': Significant wave height.
        - 'Tp': Peak period.
        - 'gamma': Peak enhancement factor.
        - 'p0': Initial parameter values used for fitting.

    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """
    if initial_values is None:
        initial_values = {}
        
    Hs0 = np.sqrt(np.trapz(S, w))*4

    p0 = {'Hs': Hs0, 'Tp': 1, 'gamma': 1}     #standard values of curve_fit are 1
    p0.update(**initial_values)

    if w[0]==0:
        w = w[1:]
        S = S[1:]

    fun = lambda om, Hs, Tp, gamma: jonswap(Hs, Tp, gamma, sigma=sigma)(om)
    popt,__ = curve_fit(fun, w, S)
    out = dict(Hs=popt[0], Tp=popt[1], gamma=popt[2], p0=[p0['Hs'], p0['Tp'], p0['gamma']])

    return out

def onesided_to_twosided(omega, S, axis=-1):
    """
    Convert a one-sided spectrum to a two-sided spectrum.

    Parameters
    ----------
    omega : array_like
        One-sided angular frequency array.
    S : array_like
        One-sided spectral density array.
    axis : int, optional
        Axis along which to concatenate. Default is -1.

    Returns
    -------
    omega2 : ndarray
        Two-sided angular frequency array.
    S2 : ndarray
        Two-sided spectral density array.

    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """
    S2 = 0.5*np.concatenate([np.flip(S, axis=axis), S], axis=axis)
    omega2 = np.hstack([np.flip(-omega), omega])
    
    return omega2, S2

def twosided_to_onesided(omega, S):
    """
    Convert a two-sided spectrum to a one-sided spectrum.

    Parameters
    ----------
    omega : array_like
        Two-sided angular frequency array.
    S : array_like
        Two-sided spectral density array.

    Returns
    -------
    omega1 : ndarray
        One-sided angular frequency array.
    S1 : ndarray
        One-sided spectral density array.

    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """
    n_samples = len(omega)    
    return omega[:n_samples//2], S[:,:,:n_samples//2]

def ramp_up(Nramp, Ntot):
    """
    Create a ramp-up scaling array.

    Parameters
    ----------
    Nramp : int
        Number of samples for ramp-up.
    Ntot : int
        Total number of samples.

    Returns
    -------
    t_scale : ndarray
        Array with ramp-up scaling applied.

    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """
    t_scale = np.ones(Ntot)
    t_scale[:Nramp] = np.linspace(0, 1, Nramp)
    return t_scale

def ramp_up_t(t, t0):
    """
    Create a ramp-up scaling array based on time threshold.

    Parameters
    ----------
    t : array_like
        Time array.
    t0 : float
        Time threshold for ramp-up.

    Returns
    -------
    t_scale : ndarray
        Array with ramp-up scaling applied.

    Notes
    -----
    Docstring is generated by GitHub Copilot.
    """
    Nramp = np.sum(t<t0)
    Ntot = len(t)
    
    return ramp_up(Nramp, Ntot)


# -*- coding: utf-8 -*-
import numpy as np
import re
import json
import csv
import dill
from pathlib import Path

from wawi.model import ModalDry, PontoonType, Node, Model, Aero, AeroSection

def import_folder(model_folder, pontoon_stl='pontoon.stl', aero_sections='aero_sections.json',
                modal='modal.json', pontoon='pontoon.json', eldef='element.json', orientations='orientations.json',
                drag_elements='drag_elements.json', pontoon_types='pontoon_types.json', 
                sort_nodes_by_x=True, interpolation_kind='linear'):
    
    '''
    Import folder containing files defining a WAWI model.

    Parameters
    ---------------
    model_folder : str
        string defining path of root
    pontoon_stl : 'pontoon.stl', optional
        string to relative path of stl-file with pontoon (3d file)
    aero_sections : 'aero_sections.json', optional
        string to relative path of json file with aerodynamic sections
    modal : 'modal.json', optional
        string to relative path of dry modal definitions
    pontoon : 'pontoon.json', optional
        string to relative path of pontoon definitions
    eldef : 'element.json', optional
        string to relative path of element definitions (node and element matrices)
    orientations : 'orientations.json', optional
        string to relative path of orientations of aero elements
    drag_elements : 'drag_elements.json', optional
        string to relative path of file defining drag elements
    pontoon_types : 'pontoon_types.json', optional
        string to relative path of pontoon types
    sort_nodes_by_x : True, optional
        whether or not to sort the nodes of eldef by their x-coordinate
    interpolation_kind : {'quadratic', 'linear', ...}
        interpolation kind used for hydrodynamic transfer function and hydrodynamic system matrices

    Returns
    ---------------
    model : `wawi.model.Model`
        WAWI model object

    Notes
    ---------------
    Often, it is more convenient to define all (or many) of the model parameters in input files rather than in scripts. To simplify this process, a support function `wawi.io.import_folder` is provided in WAWI. An example of how to use the function is given in the example [Folder import of models.ipynb](./examples/0 Model generation and setup/JSON definition and folder import/Folder import of models.ipynb)

    This imports all files in the specified folder. The file names of the relevant files in the provided path are defined as inputs to the function (default values in parentheses): 

    - modal ('modal.json')
    - pontoon('pontoon.json')
    - pontoon_types ('pontoon_types.json')
    - pontoon_stl ('pontoon.stl')
    - eldef ('element.json')
    - aero_sections ('aero_sections.json')
    - orientations ('orientations.json')
    - drag_elements ('drag_elements.json')

    Modal definition (modal)
    ^^^^^^^^^^^^^^^^^^^^^^^^
    The first file, 'modal.json', defines the dry modes of the structure under consideration. It includes the following fields:

    ```json
    {
      "local": false,
      "phi": {
        "full": [[...]],
        "hydro": [[...]],
        "girder": [[...]],
        "girder_forces": [[...]]
      },
      "m": [...],
      "omega_n": [...]
    }
    ```

    The `local` parameter defines if the modal transformation matrices (mode shapes) are given in local or global coordinate system. Typically, a global system is more convenient, but both types should work fine in WAWI. The `phi` parameter gives a dictionary defining collections of mode shapes (or modal transformation matrices). The values should be list of lists (inner list is a row) describing the modal transformation matrix of a given key. The matrices have dimensions dofs-by-modes. The keys 'full' and 'hydro' are defaults in WAWI for the full model (used for plotting and convenient for aerodynamic analysis) and pontoon loading, respectively (they can be modified, but sticking with the defaults is a good choice). Furthermore, at least two of the following modal parameters must be given: `m` (modal mass) and `omega_n` (undamped natural frequencies in rad/s) or `m` and `k` (modal stiffness). All of these are 1d lists with length matching the number of columns in all entries in `phi`.

    Pontoon and pontoon types (pontoon, pontoon_types, pontoon_stl)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    All pontoon-related definitions are given by the files `pontoon.json`, `pontoon_types.json` and `pontoon.stl`. 

    Firstly, `pontoon.json` defines pontoon coordinates, orientations and pontoon type definition, as follows:

    ```json
    {
      "P1": {
        "coordinates": [0.0, 0.0, 0.0],
        "pontoon_type": "ptype_1",
        "rotation": 0,
        "node": 1
      },
      "P2": {
        "coordinates": [100.0, 0.0, 0.0],
        "pontoon_type": "ptype_1",
        "rotation": 12.5,
        "node": 2
      }
    }
    ```

    This defines two pontoons (named P1 and P2), places them at coordinates (0,0,0) and (100.0, 0, 0) with orientation (0 deg and 12.5 deg), assigns pontoon type from files "ptype_1.npz". Also, the `node` parameter connects the pontoons to nodes 1 and 2, which should be present in the node matrix of the full system (see `node_matrix` in `element.json`). When a pontoon is connected to a node, the pontoon is automatically plotted together with the elements when using the plotting methods of the model (both rotation and location of the node is used). Also, the coordinates of the pontoon is overwritten by the coordinates of the node. Only the x and y coordinates of the pontoon matters for the computation, as the resulting wave excitation from irregular wave fields is dependent on the horizontal coordinates. The pontoon type definition is an npz-file containing the following numpy nd arrays:

    * Added mass: 'M' (6 x 6 x len(omega))
    * Radiation/added damping: 'C' (6 x 6 x len(omega))
    * Linear hydrodynamic transfer function matrix: 'Q' (6 x len(theta_Q) x len(omega_Q))
    * omega (for mass and damping): 'omega' 
    * theta for transfer: 'theta_Q'
    * omega for transfer: 'omega_Q'

    Second, the file `pontoon_types.json` is usually not needed. It is merely added to enable adding additional properties to the pontoon types, not present in the npz-files. Currently, this is limited to drag coefficients and drag areas for inclusion of linearized drag effects (stochastic linearization). It could look something like this (drag terms ordered as relevant to x, y and z-directions, respectively):

    ```json
    {
        "ptype_1": {
            "Cd": [1.0, 1.0, 2.0],
            "area": [1.0, 2.0, 2.0]
        },

        "ptype_2": {
            "Cd": [0.5, 0.4, 0.8],
            "area": [1.0, 2.0, 2.0]
        }
    }
    ```

    If instead a wildcard name is used, these properties are used for all pontoons:

    ```json
    {
        "*": {
            "Cd": [1.0, 1.0, 2.0],
            "area": [1.0, 2.0, 2.0]
        }
    }
    ```

    Finally, the `pontoon.stl` file is simply a step file containing the mesh used to visualize pontoons together with the elements in the plotting methods.

    Element definition (eldef)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^
    The file `element.json` defines the elements of the model (for aerodynamic analysis and plotting). It must contain the keys 'element_matrix' and 'node_matrix', and can also contain the fields 'sections' and 'section_assignments'. The two first are list of lists where the innermost lists are rows. These follow the typical convention in input files; rows of the node matrix consist of node label, x-coordinate, y-coordinate and z-coordinate, whereas the rows of the element matrix consist of the element label, label of the first node and label of the second node.

    The two latter properties of the files are not used inside WAWI directly (unless using drag elements, see below), but are used for the definition of the elements in the BEEF environment. It can therefore be useful for if the user for instance wants to calculate the stresses in an element, as all relevant properties are present inside the `eldef` attribute of the `model` object. The `sections` property is assumed self-explanatory from the json example provided. The `section_assignment` property is built up with a list of strings where each entry corresponds to the same-index row of the element matrix.

    ```json
    {
        "element_matrix": [[...]],
        "node_matrix":[[...]],

        "sections": 
        {
            "girder":
                {"E": 210000000000.0, 
                "Iz": 0.09285714285714286, 
                "Iy": 0.09285714285714286, 
                "A": 0.0052928571428571426, 
                "poisson": 0.3, 
                "J": 0.10714285714285714, 
                "m": 45.6272019496974},

            "column":
                {"E": 210000000000.0, 
                "Iz": 0.09285714285714286, 
                "Iy": 0.09285714285714286, 
                "A": 0.0052928571428571426, 
                "poisson": 0.3, 
                "J": 0.10714285714285714, 
                "m": 45.6272019496974}
        },

        "section_assignments": ["girder", "girder", "girder", "column", "column"]
    }
    ```

    Aerodynamic sections (aero_sections, orientations)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    Aerodynamic sections relate aerodynamic properties to the elements defined in the section above. Two definitions are required, and are given in the files
    `aero_sections.json` and `orientations.json`.

    First, the aerodynamic sections are defined in the `aero_sections.json`-file, as follows:

    ```json
    {
        "girder":
            {
                "B":26.8,
                "D":3.715,

                "Cd":0.679,
                "dCd":-1.022,
                "Cl":-0.3896,
                "dCl":4.0116,
                "Cm":-0.0700,
                "dCm":1.0122,

                "elements": [1000, 1001, 1002, 1003, ...]
            },

        "column":
        {
            "B":3,
            "D":3,

            "Cd":0.679,
            "dCd":-1.022,
            "Cl":-0.3896,
            "dCl":4.0116,
            "Cm":-0.0700,
            "dCm":1.0122,

            "elements": [2001, 2002, 2003, 2004, 2005, ...]
        }
    }
    ```

    Here, two aerodynamic sections are defined. This includes definitions of width (B), height (D), static wind coefficients (Cd, dCd, Cl, dCl, Cm, and dCm - see `wawi.model.aero.AeroSection` for details on this) and a list of elements to apply the sections to.

    For aerodynamic analysis, the orientations of the elements are crucial. These are defined in the second file (`orientations.json`). The following structure is used for this:

    ```json
    {
        "girder": {
            "e3": [0,0,1],
            "elements": [1000, 1001, 1002, 1003, ...]
        },
        "column": {
            "e2": [0,1,0],
            "elements": [2001, 2002, 2003, 2004, 2005, ...]
        }
    }
    ```

    In the above snippet, two definitions of orientations are given. Note that the keys here are not used for anything (and are not related to the `aero_sections.json` keys); they are merely included to make it convenient for the user to maintain a structured model setup. The orientation is given by either `e2` or `e3`, indicating if the y/drag or z/lift of the element is defined. Note also that these definitions are assumed approximate; the procedure inside WAWI consists of taking a cross product between the supplied unit vector and the axial direction of the element, which gives a proper perpendicular vector. The last perpendicular vector is thereafter based on a secondary cross product between the calculated perpendicular unit vector and the longitudinal vector. Finally, a list of elements to assign the orientation to is provided. The resulting orientation is conveniently plotted in the model using `model.plot(tmat_on=['undeformed'], tmat_scaling=100)` (adjust `scaling` to make sense for your problem).

    Drag elements (drag_elements)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    To accommodate the possibility for floating bridges to have mooring lines with significant quadratic drag damping, an additional file for defining these properties is supported. Note that this, as for the drag of the pontoons, is only solved in a linearized sense using (approximate) stochastic linearization techniques. The `drag_elements.json` file is defined as follows:

    ```json
    {
        "mooring_wire":
            {   "sections": "MooringWire",
                "zrange": [-1e4, 0.0],
                "rho": 1025.0,
                "Cd": [0.1, 1.2, 1.2],
                "D": 0.146}
    }
    ```

    For the drag element group named "mooring_wire" above, all elements present in the elements assigned with the relevant section assignment(s) in `element.json` (either list of several strings or single string defined with property `sections`) and that have a z-coordinate inside the provided `zrange` will be given quadratic drag damping according to the specified density (`rho`), drag coefficients in local x, y and z-direction (`Cd`) and diameter (`D`).
    '''


    from beef.fe import Part, Section
    from beef.general import n2d_ix
    
    model_folder = Path(model_folder)
    
    # Load pontoon data
    pontoon_data = {}
    if pontoon is not None:
        try:
            with open(model_folder / pontoon, 'r') as f:
                pontoon_data = json.load(f)
        except:
            print('Valid pontoon data not found. No pontoons created.')
           

    # Load modal data
    with open(model_folder / modal, 'r') as f:
        modal = json.load(f)
    
    # Load element definitions
    if eldef is not None:
        try:
            with open(model_folder / eldef, 'r') as f:
                element_data = json.load(f)
        except:
            print('Valid element data not found. No eldef created.')
            element_data = {}
    else:
        element_data = {}
            
    if orientations is not None:
        try:
            with open(model_folder / orientations, 'r') as f:
                orientations = json.load(f)
        except:
            print('Specified orientations file found. No orientation definitions applied.')
            orientations = {}
    else:
        orientations = {}


    common_ptype_settings = None
    if pontoon_types is not None:
        try:
            with open(model_folder / pontoon_types, 'r') as f:
                pontoon_type_settings = json.load(f)
            
            if '*' in pontoon_type_settings:
                common_ptype_settings = pontoon_type_settings.pop('*')
            
        except:
            print('Valid pontoon type settings file not found. No definitions applied.')
            pontoon_type_settings = {}

    else:
        pontoon_type_settings = {}
    
    # Extract specific pontoon data
    pontoon_names = [key for key in pontoon_data]
    pontoon_nodes = [Node(pontoon_data[key]['node'], *pontoon_data[key]['coordinates']) for key in pontoon_data]
    pontoon_rotation = [pontoon_data[key]['rotation']*np.pi/180 for key in pontoon_data]
    pontoon_types = [pontoon_data[key]['pontoon_type'] for key in pontoon_data]
   
    # Pontoon types
    unique_pontoon_types = list(set(pontoon_types))
    files = [f'{model_folder}/{pt}.npz' for pt in unique_pontoon_types]
    ptypes = dict()    
    
    if len(pontoon_type_settings)==0:    
        pontoon_type_settings = {key: {} for key in unique_pontoon_types}
    
    if common_ptype_settings: #universal setting, overwriting others
        for name in pontoon_type_settings:
            for par in common_ptype_settings:
                pontoon_type_settings[name][par] = common_ptype_settings[par]
    
    for file in files:
        name = Path(file).stem
        P = np.load(file)
        
        if 'Cd' in P:
            Cd = P['Cd']
        else:
            Cd = np.array([0,0,0,0,0,0])
        
        if 'A' in P:
            A = P['A']
        else:
            A = np.array([0,0,0,0,0,0])
            
        ACd = {'area': A, 'Cd':Cd}
        ptypes[name] = PontoonType.from_numeric(interpolation_kind=interpolation_kind, A=P['M'], B=P['C'], 
                         omega=P['omega'], Q=P['Q'], theta=P['theta_Q'], omegaQ=P['omega_Q'], label=name, 
                         stl_path=f'{model_folder}/{pontoon_stl}', **{**ACd, **pontoon_type_settings[name]})    

    
    # Modal dry object 
    if 'xi0' in modal:
        xi0 = np.array(modal.pop('xi0'))
    else:
        xi0 = 0.0
        
    if 'm_min' in modal:
        m_min = modal.pop('m_min')
    else:
        m_min = 0.0
        
    if 'phi_x' in modal:
        phi_x = modal.pop('phi_x')
    else:
        phi_x = None
    
    if 'local' in modal:
        local_phi = modal.pop('local')
    else:
        local_phi = False    
    
    phi = modal.pop('phi')

    modal_dry = ModalDry(phi, xi0=xi0, local_phi=local_phi, phi_x=phi_x, m_min=m_min, **modal)
    
    # Element definition
    if element_data != {}:
        if 'sections' in element_data:
            sd = element_data['sections']
            section_dict = {key: Section(E=sd[key]['E'], poisson=sd[key]['poisson'], 
                                          A=sd[key]['A'], I_y=sd[key]['Iy'], I_z=sd[key]['Iz'], 
                                          J=sd[key]['J'], m=sd[key]['m'], name=key) for key in sd}
    
            sections = [section_dict[key] for key in element_data['section_assignment']]
        else:
            sections = None
            
        node_matrix = np.array(element_data['node_matrix'])
        node_matrix[:,0] = node_matrix[:,0].astype(int)
        element_matrix = np.array(element_data['element_matrix'])
        element_matrix = element_matrix.astype(int)

        # Remove elements without valid nodes
        remove_ix = []
        remove_els = []

        for ix,row in enumerate(element_matrix):
            el,node1,node2 = row
            # print(row)

            if (node1 not in node_matrix[:,0].astype(int)) or (node2 not in node_matrix[:,0].astype(int)):
               remove_ix.append(ix) 
               remove_els.append(el)

        if len(remove_els)>0:
            print(f'Elements {remove_els} do not have valid nodes - not included in model.')
            
        element_matrix = np.delete(element_matrix, remove_ix, axis=0)            

        eldef = Part(node_matrix, element_matrix, sections=sections,
                     assemble=False, forced_ndofs=6)
            
        if sort_nodes_by_x:
            ix = np.argsort([n.coordinates[0] for n in eldef.nodes])
            node_labels_sorted = eldef.get_node_labels()[ix]
            eldef.arrange_nodes(node_labels_sorted, arrange_dof_ixs=True)
              
            if 'full' in modal_dry.phi_full:  #adjust phi_full
                modal_dry.phi_full['full'] = modal_dry.phi_full['full'][n2d_ix(ix, n_dofs=6), :]

        # Assign orientation
        for key in orientations:
            if 'e2' in orientations[key]:
                e2 = orientations[key]['e2']
                e3 = None
            elif 'e3' in orientations[key]:
                e3 = orientations[key]['e3']
                e2 = None               
            else:
                raise ValueError('Orientations should contain either e2 or e3')

            elements = orientations[key]['elements']
            for el_label in elements:
                el = eldef.get_element(int(el_label))

                el.assign_e2(e2)
                el.assign_e3(e3)

                el.initiate_geometry()
    else:
        eldef = None

    # Create model object (only hydro part)
    model = Model.from_nodes_and_types(pontoon_nodes, [ptypes[pt] for pt in pontoon_types], modal_dry=modal_dry, 
                                       rotation=pontoon_rotation, eldef=eldef, labels=pontoon_names)
    
    # Aero mode
    if aero_sections is not None:
        # Load aero sections
        try:
            with open(model_folder / aero_sections, 'r') as f:
                data = json.load(f)
        
            sections = dict()
            element_assignments = dict()
            for key in data:
                element_assignments[key] = data[key].pop('elements')
                sections[key] = AeroSection(**data[key])
            
            model.aero = Aero(sections=sections, element_assignments=element_assignments)
        except:
            print('Specified aero_sections file found. No aerodynamics definitions applied.')

    # Drag elements model
    if drag_elements is not None:
        try:
            with open(model_folder / drag_elements, 'r') as f:
                data = json.load(f)
            
            model.assign_drag_elements(data)
        except:
            print('Specified drag_elements file found or invalid. No drag elements defined.')
    
    model.connect_eldef()
    model.assign_dry_modes()
    
    return model


def save_wwi(model, model_path):
    '''
    Save WAWI model object to specified model path.
    '''
    with open(model_path, 'wb') as f:
        dill.dump(model, f, -1)
    

def convert_to_wwi(model_path, output=None):
    '''
    Convert folder with WAWI object definition files to wwi (model) file
    '''
    model_path = Path(model_path)
    if output is None:
        output = model_path / 'model.wwi'
    
    model = load_model(model_path)

    if model_path.is_dir():
        with open(output, 'wb') as f:
            dill.dump(model, f, -1)


def load_model(model_path, save_if_nonexistent=False, sort_nodes_by_x=True):
    '''
    Load wwi-file with WAWI model object or folder with relevant definitions.

    Parameters
    -----------
    model_path : str
        path of wwi-file
    save_if_nonexistent : False, optional
        whether or not to save model as wwi if not existing
    sort_nodes_by_x : True, optional
        whether or not to sort the nodes of eldef by their x-coordinate

    Returns
    -----------
    model : `wawi.model.Model` 
      WAWI model
    '''

    model_path = Path(model_path)

    if model_path.is_dir() and not (model_path/'model.wwi').is_file():
        model = import_folder(model_path, sort_nodes_by_x=sort_nodes_by_x)
        if save_if_nonexistent:
            convert_to_wwi(model_path)   
    else:
        with open(model_path, 'rb') as f:
            model = dill.load(f)
    
    return model
            

def import_wadam_hydro_transfer(wadam_file):
    '''
    Import WADAM LIS output file.

    Parameters
    -----------
    wadam_file : str
        path to file to import
    
    Returns
    ----------
    omega : float
        numpy array describing numerical frequency axis
    theta : float
        numpy array describing numerical directional axis
    fhyd : float
        complex numpy 3d-array describing the transfer functions relating
        regular waves with unit height to the 6 relevant forces and moments
        6-by-len(theta)-by-len(omega)
    '''

    string = (r'.+W A V E  P E R I O D.+=\s+(?P<period>.+):.+\n.+'
              r'H E A D I N G   A N G L E.+=\s+(?P<theta>.+):(?:.*\n){5,10}'
              
              r'.+EXCITING FORCES AND MOMENTS FROM THE HASKIN RELATIONS(?:.*\n){4,6}\s+'
              
              r'-F1-\s+(?P<F1_real>[-?.E\d+]+)\s+(?P<F1_imag>[-?.E\d+]+)\s.+\n\s+'
              r'-F2-\s+(?P<F2_real>[-?.E\d+]+)\s+(?P<F2_imag>[-?.E\d+]+)\s.+\n\s+'
              r'-F3-\s+(?P<F3_real>[-?.E\d+]+)\s+(?P<F3_imag>[-?.E\d+]+)\s.+\n\s+'
              r'-F4-\s+(?P<F4_real>[-?.E\d+]+)\s+(?P<F4_imag>[-?.E\d+]+)\s.+\n\s+'
              r'-F5-\s+(?P<F5_real>[-?.E\d+]+)\s+(?P<F5_imag>[-?.E\d+]+)\s.+\n\s+'
              r'-F6-\s+(?P<F6_real>[-?.E\d+]+)\s+(?P<F6_imag>[-?.E\d+]+)\s.+\n')
    regex = re.compile(string) 
    
    nondim_string = (r'\s+NON-DIMENSIONALIZING FACTORS:\n(?:.*\n)+'
                     r'\s+RO\s+=\s+(?P<rho>[-?.E\d+]+)\n'
                     r'\s+G\s+=\s+(?P<g>[-?.E\d+]+)\n'
                     r'\s+VOL\s+=\s+(?P<vol>[-?.E\d+]+)\n'
                     r'\s+L\s+=\s+(?P<l>[-?.E\d+]+)\n'
                     r'\s+WA\s+=\s+(?P<wa>[-?.E\d+]+)')
    
    regex_nondim = re.compile(nondim_string) 
    
    with open(wadam_file, encoding='utf-8') as file:
        data = file.read()
    
    periods = []
    thetas = []
    Q = []
    
    for match in regex.finditer(data):
        data_dict = dict(zip(match.groupdict().keys(), [float(val) for val in match.groupdict().values()]))
        F1 = data_dict['F1_real'] + data_dict['F1_imag']*1j
        F2 = data_dict['F2_real'] + data_dict['F2_imag']*1j
        F3 = data_dict['F3_real'] + data_dict['F3_imag']*1j
        F4 = data_dict['F4_real'] + data_dict['F4_imag']*1j
        F5 = data_dict['F5_real'] + data_dict['F5_imag']*1j
        F6 = data_dict['F6_real'] + data_dict['F6_imag']*1j
        Q.append(np.array([F1,F2,F3,F4,F5,F6]))
        
        thetas.append(data_dict['theta'])
        periods.append(data_dict['period'])
    
    theta = np.unique(np.array(thetas))*np.pi/180.0
    period = np.unique(np.array(periods))
    omega = np.flip(2*np.pi/period, axis=0)
    Q = np.vstack(Q).T
    fhyd = np.zeros([6, len(theta), len(omega)]).astype('complex')
    for n in range(len(theta)):
        for k in range(len(omega)):
            fhyd[:,n,k] = Q[:, k*len(theta)+n] 
    
    # Re-dimensionalize and flip
    nondim_parameters = regex_nondim.search(data).groupdict()
    nd = dict(zip(nondim_parameters.keys(), [float(val) for val in nondim_parameters.values()]))
    
    rho, g, vol, wa, l = nd['rho'], nd['g'], nd['vol'], nd['wa'], nd['l']
    dim = 0*fhyd
    dim[0::6,:,:] = rho*vol*g*wa/l
    dim[1::6,:,:] = rho*vol*g*wa/l
    dim[2::6,:,:] = rho*vol*g*wa/l
    
    dim[3::6,:,:] = rho*vol*g*wa
    dim[4::6,:,:] = rho*vol*g*wa
    dim[5::6,:,:] = rho*vol*g*wa

    fhyd = np.flip(fhyd, axis=2) * dim

    return omega, theta, fhyd


def import_wadam_mat(wadam_file):
    '''
    Import system matrices given by input WADAM results (frequency dependent and constant).

    Parameters
    -------------
    wadam_file : str
        path to WADAM file to open
    
    Returns
    -------------
    mass : float
        6-by-6 added mass matrix of body
    damping : float
        6-by-6 radiation damping matrix of body
    stiffness : float
        6-by-6 restoring stiffness matrix of body
    static_mass : float
        6-by-6 static (constant) mass matrix of body
        this represents inertia of pontoon object itself
    omega : float
        numpy array describing numerical frequency axis (corresponding to mass and damping
        output)
    '''

    static_mass_target = 'MASS INERTIA COEFFICIENT MATRIX'
    stiffness_target = 'HYDROSTATIC RESTORING COEFFICIENT MATRIX'
    mass_target = 'ADDED MASS MATRIX                                   '
    damping_target = 'DAMPING MATRIX                                      '
    period_target = 'WAVE PERIOD  =   '
    non_dim_target = ' THE OUTPUT IS NON-DIMENSIONALIZED USING -'

    f = open(wadam_file)
    active_search = False
    current_elements = []
    current_range = [0]

    stiffness = np.empty([0,6,6])
    static_mass = np.empty([0,6,6])
    mass = np.empty([0,6,6])
    damping = np.empty([0,6,6])
    period = np.empty([0,1,1])
    non_dim = np.empty([0])

    temp_data=[]
    data_storage = {'static_mass': static_mass, 'stiffness': stiffness, 'period': period, 'mass':mass,'damping':damping,'non_dim':non_dim}
    append_axis = {'static_mass': 0, 'stiffness': 0, 'period': 0, 'mass':0,'damping':0,'non_dim':None}

    for lineno, line in enumerate(f):
        static_mass_switch = line.find(static_mass_target)!=-1
        stiffness_switch = line.find(stiffness_target)!=-1
        period_switch = line.find(period_target)!=-1
        mass_switch = line.find(mass_target)!=-1
        damping_switch = line.find(damping_target)!=-1
        non_dim_switch = line.find(non_dim_target)!=-1

        if static_mass_switch == True:
            active_search=True
            current_range = [el+lineno-1 for el in range(5,11)]
            current_elements = range(1,6+1)
            current_data_string='static_mass'
            current_data = data_storage[current_data_string]
        elif stiffness_switch:
            active_search=True
            current_range = [el+lineno-1 for el in range(5,11)]
            current_elements = range(1,6+1)
            current_data_string='stiffness'
            current_data = data_storage[current_data_string]
        elif period_switch:
             active_search=True
             current_range = [lineno]
             current_elements = [3]
             current_data_string='period'
             current_data = data_storage[current_data_string]
        elif mass_switch:
            active_search=True
            current_range = [el+lineno-1 for el in range(5,11)]
            current_elements = range(1,6+1)
            current_data_string='mass'
            current_data = data_storage[current_data_string]
        elif damping_switch:
            active_search=True
            current_range = [el+lineno-1 for el in range(5,11)]
            current_elements = range(1,6+1)
            current_data_string='damping'
            current_data = data_storage[current_data_string]
        elif non_dim_switch:
            active_search=True
            current_range = [el+lineno-1 for el in range(9,14)]
            current_elements = [2]
            current_data_string='non_dim'
            current_data = data_storage[current_data_string]

        if (active_search==True) and (lineno in current_range):
            data_line = [float(i) for i in [line.split()[el] for el in current_elements]]
            temp_data.append(data_line)
        elif (active_search==True) and lineno>=max(current_range):
             current_data=np.append(current_data,[temp_data],axis=append_axis[current_data_string])
             temp_data=[]
             current_range = [0]
             current_elements = []
             active_search=False
             data_storage[current_data_string] = current_data

    # Non-dimensionalizing factors
    non_dim = data_storage['non_dim']
    ro = non_dim[0]
    g = non_dim[1]
    vol = non_dim[2]
    l = non_dim[3]

    non_dim_defs = {'static_mass': np.array([[ro*vol, ro*vol*l],
                                             [ro*vol*l, ro*vol*l*l]]), 
                    'mass': np.array([[ro*vol, ro*vol*l],
                                      [ro*vol*l, ro*vol*l*l]]), 
                    'damping': np.array([[ro*vol*np.sqrt(g/l), ro*vol*np.sqrt(g*l)],
                                         [ro*vol*np.sqrt(g*l), ro*vol*l*np.sqrt(g*l)]]), 
                    'stiffness':np.array([[ro*vol*g/l, ro*vol*g],
                                          [ro*vol*g, ro*vol*g*l]])}

    # Re-dimensionalize data
    for current_data_string in non_dim_defs.keys():
        dim_matrix = np.repeat(np.repeat(non_dim_defs[current_data_string],3,axis=0),3,axis=1)
        data_storage[current_data_string] = np.multiply(data_storage[current_data_string], dim_matrix)

    static_mass = data_storage['static_mass'][0]
    stiffness = data_storage['stiffness'][0]
    mass = np.moveaxis(data_storage['mass'],0,2)
    damping = np.moveaxis(data_storage['damping'],0,2)
    omega = 2.0*np.pi/data_storage['period'].flatten()
    sortix = np.argsort(omega)
    omega = omega[sortix]
    
    mass = mass[:,:, sortix]
    damping = damping[:,:, sortix]
    stiffness = (stiffness + stiffness.T)/2

    return  mass, damping, stiffness, static_mass, omega



def import_wamit_mat(wamit_path, suffix=['hst', 'mmx', '1'], rho=1020, L=1.0, g=9.80665):
    '''
    Import system matrices from WAMIT results (frequency dependent and constant).

    Parameters
    -------------
    wamit_path : str
        path to WAMIT file to open
    suffix : ['hst', 'mmx', '1']
        all suffixes used - all default values required for full output
    rho : 1020, optional
        water mass density used to redimensionalize results
    L : 1.0, optional
        redimensionaling length - 1.0 is fine for most use cases
    g : 9.80665, optional
        gravitational constant for redimensionalization of data
    
    Returns
    -------------
    A : float
        6-by-6 added mass matrix of body
    B : float
        6-by-6 radiation damping matrix of body
    K0 : float
        6-by-6 restoring stiffness matrix of body
    M0 : float
        6-by-6 static (constant) mass matrix of body
        this represents inertia of pontoon object itself
    omega : float
        numpy array describing numerical frequency axis (corresponding to mass and damping
        output)
    '''
    if suffix.count('hst')>=1:
        # Hydrostatic stiffness
        k_list = np.loadtxt(wamit_path+'.hst')
        K0 = np.reshape(k_list[:,2], [6,6])   #assumes the order of elements in k_list
        dimK = 0*K0
        dimK[2, 2] = rho*g*L**2
        dimK[2, 3] = dimK[3, 2] = rho*g*L**3
        dimK[2, 4] = dimK[4, 2] = rho*g*L**3
        dimK[3::, :]  = rho*g*L**4
        dimK[:, 3::]  = rho*g*L**4
        
        K0 = K0*dimK
        
    else:
        K0 = 0

    if suffix.count('mmx')>=1:
        # Inertia matrix
        m_list = np.loadtxt(wamit_path+'.mmx', skiprows=12)
        M0 = np.reshape(m_list[:,2], [6,6]) 
    else: 
        M0 = 0

    if suffix.count('1')>=1:
        # Added mass and added damping
        mc_list = np.loadtxt(wamit_path+'.1')
        period = mc_list[0::10,0]
        Aper = np.zeros([6,6,len(period)])
        Bper = np.zeros([6,6,len(period)])

        for k in range(0,int(mc_list.shape[0]/10)):
            mc_this = mc_list[k*10:k*10+10,:]
            dof1, dof2 = mc_this[:,1].astype('int')-1, mc_this[:, 2].astype('int')-1

            for ix, i in enumerate(dof1):
                j = dof2[ix]
                Aper[i, j, k] = mc_this[ix, 3] 
                Bper[i, j, k] = mc_this[ix, 4] 

        # Re-dimensionalize and flip
        omega = np.flip(2*np.pi/period, axis=0)

        dimA = 0*Aper
        dimA[0:3, 0:3, :] = rho*L**3
        dimA[0:3, 3:-1, :] = rho*L**4
        dimA[3:-1, 0:3, :] = rho*L**4
        dimA[3:-1, 3:-1, :] = rho*L**5

        dimB = 0*Bper
        dimB[0:3, 0:3, :] = rho*L**3*omega
        dimB[0:3, 3:-1, :] = rho*L**4*omega
        dimB[3:-1, 0:3, :] = rho*L**4*omega
        dimB[3:-1, 3:-1, :] = rho*L**5*omega
        
        A = np.flip(Aper, axis=2)*dimA
        B = np.flip(Bper, axis=2)*dimB
    else:
        A = 0
        B = 0

    return A, B, K0, M0, omega


def import_wamit_force(wamit_path, A=1.0, rho=1020, L=1.0, g=9.80665):
    '''
    Import hydrodynamic transfer function from WAMIT '.2'-file.

    Parameters
    -----------
    wamit_path :
    A : 1.0, optional
        area used to redimensionalize results (1.0 is fine for most use cases)
    rho : 1020, optional
        water mass density used to redimensionalize results
    L : 1.0, optional
        redimensionaling length  (1.0 is fine for most use cases)
    g : 9.80665, optional
        gravitational constant for redimensionalization of data

    Returns
    ----------
    omega : float
        numpy array describing numerical frequency axis
    theta : float
        numpy array describing numerical directional axis
    fhyd : float
        complex numpy 3d-array describing the transfer functions relating
        regular waves with unit height to the 6 relevant forces and moments
        6-by-len(theta)-by-len(omega)
    '''

    # Transfer function
    q_list = np.loadtxt(wamit_path+'.2')    
    period = np.unique(q_list[:,0])
    n_omega = len(period)
    n_theta = int(len(q_list[:,0])/len(period)/6)
    theta = q_list[0:6*n_theta:6,1]
    f_hyd_per = np.zeros([6, n_theta, n_omega]).astype('complex')

    for k in range(0, n_omega):
        q_this = q_list[k*n_theta*6:k*n_theta*6+n_theta*6]
        for dof in range(0,6):
            f_hyd_per[dof, :, k] = q_this[dof::6, 5] + q_this[dof::6, 6]*1j
            
    omega = np.flip(2*np.pi/period, axis=0)
    
    # Re-dimensionalize and flip
    dim = 0*f_hyd_per
    dim[0::6,:,:] = rho*g*A*L**2
    dim[1::6,:,:] = rho*g*A*L**2
    dim[2::6,:,:] = rho*g*A*L**2
    dim[3::6,:,:] = rho*g*A*L**3
    dim[4::6,:,:] = rho*g*A*L**3
    dim[5::6,:,:] = rho*g*A*L**3

    f_hyd = np.flip(f_hyd_per, axis=2) * dim

    return omega, theta, f_hyd

